import numpy as np

import xtrack as xt

ARC_NAMES = ['12', '23', '34', '45', '56', '67', '78', '81']

def get_arc_periodic_solution(collider, line_name=None, arc_name=None):

    assert collider.lhcb1.twiss_default.get('reverse', False) is False
    assert collider.lhcb2.twiss_default['reverse'] is True

    if (line_name is None or arc_name is None
        or isinstance(line_name, (list, tuple))
        or isinstance(arc_name, (list, tuple))):

        if line_name is None:
            line_name = ['lhcb1', 'lhcb2']
        if arc_name is None:
            arc_name = ARC_NAMES

        res = {'lhcb1': {}, 'lhcb2': {}}
        for ll in line_name:
            res[ll] = {}
            for aa in arc_name:
                res[ll][aa] = get_arc_periodic_solution(
                    collider, line_name=ll, arc_name=aa)
        return res


    assert arc_name in ARC_NAMES
    assert line_name in ['lhcb1', 'lhcb2']

    beam_number = line_name[-1:]
    sector_start_number = arc_name[:1]
    sector_end_number = arc_name[1:]
    start_cell = f's.cell.{arc_name}.b{beam_number}'
    end_cell = f'e.cell.{arc_name}.b{beam_number}'
    start_arc = f'e.ds.r{sector_start_number}.b{beam_number}'
    end_arc = f's.ds.l{sector_end_number}.b{beam_number}'

    line = collider[line_name]

    twinit_cell = line.twiss(
                start=start_cell,
                end=end_cell,
                init='periodic',
                only_twiss_init=True)

    tw_to_end_arc = line.twiss(
        start=twinit_cell.element_name,
        end=end_arc,
        init=twinit_cell,
        )

    tw_to_start_arc = line.twiss(
        start=start_arc,
        end=twinit_cell.element_name,
        init=twinit_cell)

    res = xt.TwissTable.concatenate([tw_to_start_arc, tw_to_end_arc])
    res['mux'] = res['mux'] - res['mux', start_arc]
    res['muy'] = res['muy'] - res['muy', start_arc]

    return res


class ActionArcPhaseAdvanceFromCell(xt.Action):

    def __init__(self, collider, line_name, arc_name):

        self.collider = collider
        self.line_name = line_name
        self.arc_name = arc_name

    def run(self):

        tw_arc = get_arc_periodic_solution(
            self.collider, line_name=self.line_name, arc_name=self.arc_name)

        return {'table': tw_arc,
                'mux': tw_arc['mux', -1] - tw_arc['mux', 0],
                'muy': tw_arc['muy', -1] - tw_arc['muy', 0]}

def match_arc_phase_advance(collider, arc_name,
                            target_mux_b1, target_muy_b1,
                            target_mux_b2, target_muy_b2,
                            solve=True, default_tol=None):

    assert collider.lhcb1.twiss_default.get('reverse', False) is False
    assert collider.lhcb2.twiss_default['reverse'] is True

    assert arc_name in ARC_NAMES

    action_phase_b1 = ActionArcPhaseAdvanceFromCell(
                    collider=collider, line_name='lhcb1', arc_name=arc_name)
    action_phase_b2 = ActionArcPhaseAdvanceFromCell(
                        collider=collider, line_name='lhcb2', arc_name=arc_name)

    opt=collider.match(
        solve=False,
        default_tol=default_tol,
        targets=[
            action_phase_b1.target('mux', target_mux_b1),
            action_phase_b1.target('muy', target_muy_b1),
            action_phase_b2.target('mux', target_mux_b2),
            action_phase_b2.target('muy', target_muy_b2),
        ],
        vary=[
            xt.VaryList([f'kqtf.a{arc_name}b1', f'kqtd.a{arc_name}b1',
                        f'kqtf.a{arc_name}b2', f'kqtd.a{arc_name}b2',
                        f'kqf.a{arc_name}', f'kqd.a{arc_name}'
                        ]),
        ])
    if solve:
        opt.solve()

    return opt


def compute_ats_phase_advances_for_auxiliary_irs(line_name,
        tw_sq_a81_ip1_a12, tw_sq_a45_ip5_a56,
        muxip1_l, muyip1_l, muxip1_r, muyip1_r,
        muxip5_l, muyip5_l, muxip5_r, muyip5_r,
        muxip2, muyip2, muxip4, muyip4, muxip6, muyip6, muxip8, muyip8,
        mux12, muy12, mux45, muy45, mux56, muy56, mux81, muy81
    ):

    assert line_name in ['lhcb1', 'lhcb2']
    bn = line_name[-2:]

    mux_compensate_ir2 = (tw_sq_a81_ip1_a12['mux', f's.ds.l2.{bn}'] - tw_sq_a81_ip1_a12['mux', 'ip1']
                          - muxip1_r - mux12)
    mux_ir2_target = muxip2 - mux_compensate_ir2
    muy_compensate_ir2 = (tw_sq_a81_ip1_a12['muy', f's.ds.l2.{bn}'] - tw_sq_a81_ip1_a12['muy', 'ip1']
                          - muyip1_r - muy12)
    muy_ir2_target = muyip2 - muy_compensate_ir2

    mux_compensate_ir4 = (tw_sq_a45_ip5_a56['mux', 'ip5'] - tw_sq_a45_ip5_a56['mux', f'e.ds.r4.{bn}']
                          - muxip5_l - mux45)
    mux_ir4_target = muxip4 - mux_compensate_ir4
    muy_compensate_ir4 = (tw_sq_a45_ip5_a56['muy', 'ip5'] - tw_sq_a45_ip5_a56['muy', f'e.ds.r4.{bn}']
                          - muyip5_l - muy45)
    muy_ir4_target = muyip4 - muy_compensate_ir4

    mux_compensate_ir6 = (tw_sq_a45_ip5_a56['mux', f's.ds.l6.{bn}'] - tw_sq_a45_ip5_a56['mux', 'ip5']
                          - muxip5_r - mux56)
    mux_ir6_target = muxip6 - mux_compensate_ir6
    muy_compensate_ir6 = (tw_sq_a45_ip5_a56['muy', f's.ds.l6.{bn}'] - tw_sq_a45_ip5_a56['muy', 'ip5']
                          - muyip5_r - muy56)
    muy_ir6_target = muyip6 - muy_compensate_ir6

    mux_compensate_ir8 = (tw_sq_a81_ip1_a12['mux', 'ip1.l1'] - tw_sq_a81_ip1_a12['mux', f'e.ds.r8.{bn}']
                          - muxip1_l - mux81)
    mux_ir8_target = muxip8 - mux_compensate_ir8
    muy_compensate_ir8 = (tw_sq_a81_ip1_a12['muy', 'ip1.l1'] - tw_sq_a81_ip1_a12['muy', f'e.ds.r8.{bn}']
                          - muyip1_l - muy81)
    muy_ir8_target = muyip8 - muy_compensate_ir8

    return (mux_ir2_target, muy_ir2_target, mux_ir4_target, muy_ir4_target,
            mux_ir6_target, muy_ir6_target, mux_ir8_target, muy_ir8_target)


def propagate_optics_from_beta_star(collider, ip_name, line_name,
                                    beta_star_x, beta_star_y,
                                    start, end):

    assert collider.lhcb1.twiss_default.get('reverse', False) is False
    assert collider.lhcb2.twiss_default['reverse'] is True
    assert collider.lhcb1.element_names[1] == 'ip1'
    assert collider.lhcb2.element_names[1] == 'ip1.l1'
    assert collider.lhcb1.element_names[-2] == 'ip1.l1'
    assert collider.lhcb2.element_names[-2] == 'ip1'

    if ip_name == 'ip1':
        ele_stop_left = 'ip1.l1'
        ele_start_right = 'ip1'
    else:
        ele_stop_left = ip_name
        ele_start_right = ip_name

    tw_left = collider[line_name].twiss(start=start, end=ele_stop_left,
                    init=xt.TwissInit(line=collider[line_name],
                                            element_name=ele_stop_left,
                                            betx=beta_star_x,
                                            bety=beta_star_y))
    tw_right = collider[line_name].twiss(start=ele_start_right, end=end,
                        init=xt.TwissInit(line=collider[line_name],
                                                element_name=ele_start_right,
                                                betx=beta_star_x,
                                                bety=beta_star_y))

    tw_ip = xt.TwissTable.concatenate([tw_left, tw_right])

    return tw_ip

class ActionPhase_23_34(xt.Action):

    def __init__(self, collider):
        self.collider = collider

    def run(self):
        try:
            tw_arc = get_arc_periodic_solution(self.collider, arc_name=['23', '34'])
        except ValueError:
            # Twiss failed
            return {
                'mux_23_34_b1': 1e100,
                'muy_23_34_b1': 1e100,
                'mux_23_34_b2': 1e100,
                'muy_23_34_b2': 1e100,
                'mux_23_34_avg': 1e100,
                'muy_23_34_avg': 1e100,
            }
        tw_23_b1 = tw_arc['lhcb1']['23']
        tw_23_b2 = tw_arc['lhcb2']['23']
        mux_23_b1 = tw_23_b1['mux', 's.ds.l3.b1'] - tw_23_b1['mux', 'e.ds.r2.b1']
        muy_23_b1 = tw_23_b1['muy', 's.ds.l3.b1'] - tw_23_b1['muy', 'e.ds.r2.b1']
        mux_23_b2 = tw_23_b2['mux', 's.ds.l3.b2'] - tw_23_b2['mux', 'e.ds.r2.b2']
        muy_23_b2 = tw_23_b2['muy', 's.ds.l3.b2'] - tw_23_b2['muy', 'e.ds.r2.b2']

        tw34_b1 = tw_arc['lhcb1']['34']
        tw34_b2 = tw_arc['lhcb2']['34']
        mux_34_b1 = tw34_b1['mux', 's.ds.l4.b1'] - tw34_b1['mux', 'e.ds.r3.b1']
        muy_34_b1 = tw34_b1['muy', 's.ds.l4.b1'] - tw34_b1['muy', 'e.ds.r3.b1']
        mux_34_b2 = tw34_b2['mux', 's.ds.l4.b2'] - tw34_b2['mux', 'e.ds.r3.b2']
        muy_34_b2 = tw34_b2['muy', 's.ds.l4.b2'] - tw34_b2['muy', 'e.ds.r3.b2']

        return {
            'mux_23_34_b1': mux_23_b1 + mux_34_b1,
            'muy_23_34_b1': muy_23_b1 + muy_34_b1,
            'mux_23_34_b2': mux_23_b2 + mux_34_b2,
            'muy_23_34_b2': muy_23_b2 + muy_34_b2,
            'mux_23_34_avg': 0.5 * (mux_23_b1 + mux_34_b1 + mux_23_b2 + mux_34_b2),
            'muy_23_34_avg': 0.5 * (muy_23_b1 + muy_34_b1 + muy_23_b2 + muy_34_b2),
        }

class ActionPhase_67_78(xt.Action):

    def __init__(self, collider):
        self.collider = collider

    def run(self):
        try:
            tw_arc = get_arc_periodic_solution(self.collider, arc_name=['67', '78'])
        except ValueError:
            # Twiss failed
            return {
                'mux_67_78_b1': 1e100,
                'muy_67_78_b1': 1e100,
                'mux_67_78_b2': 1e100,
                'muy_67_78_b2': 1e100,
                'mux_67_78_avg': 1e100,
                'muy_67_78_avg': 1e100,
            }
        tw_67_b1 = tw_arc['lhcb1']['67']
        tw_67_b2 = tw_arc['lhcb2']['67']
        mux_67_b1 = tw_67_b1['mux', 's.ds.l7.b1'] - tw_67_b1['mux', 'e.ds.r6.b1']
        muy_67_b1 = tw_67_b1['muy', 's.ds.l7.b1'] - tw_67_b1['muy', 'e.ds.r6.b1']
        mux_67_b2 = tw_67_b2['mux', 's.ds.l7.b2'] - tw_67_b2['mux', 'e.ds.r6.b2']
        muy_67_b2 = tw_67_b2['muy', 's.ds.l7.b2'] - tw_67_b2['muy', 'e.ds.r6.b2']

        tw78_b1 = tw_arc['lhcb1']['78']
        tw78_b2 = tw_arc['lhcb2']['78']
        mux_78_b1 = tw78_b1['mux', 's.ds.l8.b1'] - tw78_b1['mux', 'e.ds.r7.b1']
        muy_78_b1 = tw78_b1['muy', 's.ds.l8.b1'] - tw78_b1['muy', 'e.ds.r7.b1']
        mux_78_b2 = tw78_b2['mux', 's.ds.l8.b2'] - tw78_b2['mux', 'e.ds.r7.b2']
        muy_78_b2 = tw78_b2['muy', 's.ds.l8.b2'] - tw78_b2['muy', 'e.ds.r7.b2']

        return {
            'mux_67_78_b1': mux_67_b1 + mux_78_b1,
            'muy_67_78_b1': muy_67_b1 + muy_78_b1,
            'mux_67_78_b2': mux_67_b2 + mux_78_b2,
            'muy_67_78_b2': muy_67_b2 + muy_78_b2,
            'mux_67_78_avg': 0.5 * (mux_67_b1 + mux_78_b1 + mux_67_b2 + mux_78_b2),
            'muy_67_78_avg': 0.5 * (muy_67_b1 + muy_78_b1 + muy_67_b2 + muy_78_b2),
        }

def change_phase_non_ats_arcs(collider,
    d_mux_15_b1=None, d_muy_15_b1=None, d_mux_15_b2=None, d_muy_15_b2=None,
    solve=True, default_tol=None):

    assert d_mux_15_b1 is not None
    assert d_muy_15_b1 is not None
    assert d_mux_15_b2 is not None
    assert d_muy_15_b2 is not None

    action_phase_23_34 = ActionPhase_23_34(collider)
    action_phase_67_78 = ActionPhase_67_78(collider)

    phase_23_34_0 = action_phase_23_34.run()
    phase_67_78_0 = action_phase_67_78.run()

    mux_23_34_b1_target = phase_23_34_0['mux_23_34_b1']
    muy_23_34_b1_target = phase_23_34_0['muy_23_34_b1']
    mux_23_34_b2_target = phase_23_34_0['mux_23_34_b2']
    muy_23_34_b2_target = phase_23_34_0['muy_23_34_b2']
    mux_67_78_b1_target = phase_67_78_0['mux_67_78_b1']
    muy_67_78_b1_target = phase_67_78_0['muy_67_78_b1']
    mux_67_78_b2_target = phase_67_78_0['mux_67_78_b2']
    muy_67_78_b2_target = phase_67_78_0['muy_67_78_b2']

    mux_23_34_b1_target += d_mux_15_b1
    mux_67_78_b1_target -= d_mux_15_b1
    muy_23_34_b1_target += d_muy_15_b1
    muy_67_78_b1_target -= d_muy_15_b1
    mux_23_34_b2_target += d_mux_15_b2
    mux_67_78_b2_target -= d_mux_15_b2
    muy_23_34_b2_target += d_muy_15_b2
    muy_67_78_b2_target -= d_muy_15_b2

    # I use match_knob to add a term instead of changing the strengths directly.
    # This is because I want to preserve the tune knobs.
    # I skip the generate_knob so that the values stay in.

    opt_mq = collider.match_knob(
        knob_name='dmu_15', knob_value_end=1,
        run=False,
        default_tol=default_tol,
        solver_options={'n_bisections': 5},
        vary=[
            xt.VaryList(['kqf.a23', 'kqd.a23', 'kqf.a34', 'kqd.a34']),
            xt.VaryList(['kqf.a67', 'kqd.a67', 'kqf.a78', 'kqd.a78']),
        ],
        targets=[
            action_phase_23_34.target('mux_23_34_avg', 0.5 * (mux_23_34_b1_target + mux_23_34_b2_target)),
            action_phase_23_34.target('muy_23_34_avg', 0.5 * (muy_23_34_b1_target + muy_23_34_b2_target)),
            action_phase_67_78.target('mux_67_78_avg', 0.5 * (mux_67_78_b1_target + mux_67_78_b2_target)),
            action_phase_67_78.target('muy_67_78_avg', 0.5 * (muy_67_78_b1_target + muy_67_78_b2_target)),
        ]
    )
    if solve:
        opt_mq.solve()

    opt_mqt_b1 = collider.match_knob(
        knob_name='dmu_15', knob_value_end=1,
        run=False,
        default_tol=default_tol,
        solver_options={'n_bisections': 5},
        vary=[
            xt.VaryList(['kqtf.a23b1', 'kqtd.a23b1', 'kqtf.a34b1', 'kqtd.a34b1']),
            xt.VaryList(['kqtf.a67b1', 'kqtd.a67b1', 'kqtf.a78b1', 'kqtd.a78b1']),
        ],
        targets=[
            action_phase_23_34.target('mux_23_34_b1', mux_23_34_b1_target),
            action_phase_23_34.target('muy_23_34_b1', muy_23_34_b1_target),
            action_phase_67_78.target('mux_67_78_b1', mux_67_78_b1_target),
            action_phase_67_78.target('muy_67_78_b1', muy_67_78_b1_target),
        ]
    )

    if solve:
        opt_mqt_b1.solve()

    opt_mqt_b2 = collider.match_knob(
        knob_name='dmu_15', knob_value_end=1,
        run=False,
        default_tol=default_tol,
        solver_options={'n_bisections': 5},
        vary=[
            xt.VaryList(['kqtf.a23b2', 'kqtd.a23b2', 'kqtf.a34b2', 'kqtd.a34b2']),
            xt.VaryList(['kqtf.a67b2', 'kqtd.a67b2', 'kqtf.a78b2', 'kqtd.a78b2']),
        ],
        targets=[
            action_phase_23_34.target('mux_23_34_b2', mux_23_34_b2_target),
            action_phase_23_34.target('muy_23_34_b2', muy_23_34_b2_target),
            action_phase_67_78.target('mux_67_78_b2', mux_67_78_b2_target),
            action_phase_67_78.target('muy_67_78_b2', muy_67_78_b2_target),
        ]
    )

    if solve:
        opt_mqt_b2.solve()

    return {'opt_mq': opt_mq, 'opt_mqt_b1': opt_mqt_b1, 'opt_mqt_b2': opt_mqt_b2}

def rematch_ir2(collider, line_name,
                boundary_conditions_left, boundary_conditions_right,
                mux_ir2, muy_ir2, betx_ip2, bety_ip2,
                solve=True, staged_match=False, default_tol=None):

    assert line_name in ['lhcb1', 'lhcb2']
    bn = line_name[-2:]

    opt = collider[f'lhc{bn}'].match(
        solve=False,
        default_tol=default_tol,
        start=f's.ds.l2.{bn}', end=f'e.ds.r2.{bn}',
        # Left boundary
        init=boundary_conditions_left, init_at=xt.START,
        targets=[
            xt.TargetSet(at=xt.END,
                    tars=('betx', 'bety', 'alfx', 'alfy', 'dx', 'dpx'),
                    value=boundary_conditions_right, tag='stage0'),
            xt.TargetSet(at='ip2',
                betx=betx_ip2, bety=bety_ip2, alfx=0, alfy=0, dx=0, dpx=0,
                tag='stage2'),
            xt.TargetRelPhaseAdvance('mux', mux_ir2, tag='stage0'),
            xt.TargetRelPhaseAdvance('muy', muy_ir2, tag='stage0'),
        ],
        vary=[
            xt.VaryList([
                f'kq9.l2{bn}', f'kq10.l2{bn}', f'kqtl11.l2{bn}', f'kqt12.l2{bn}', f'kqt13.l2{bn}',
                f'kq9.r2{bn}', f'kq10.r2{bn}', f'kqtl11.r2{bn}', f'kqt12.r2{bn}', f'kqt13.r2{bn}'],
                tag='stage0'),
            xt.VaryList(
                [f'kq4.l2{bn}', f'kq5.l2{bn}',  f'kq6.l2{bn}',    f'kq7.l2{bn}',   f'kq8.l2{bn}',
                 f'kq6.r2{bn}',    f'kq7.r2{bn}',   f'kq8.r2{bn}'],
                tag='stage1'),
            xt.VaryList([f'kq4.r2{bn}', f'kq5.r2{bn}'], tag='stage2')
        ]
    )

    if solve:
        if staged_match:
            opt.disable_all_vary()
            opt.disable_all_targets()

            opt.enable_vary(tag='stage0')
            opt.enable_targets(tag='stage0')
            opt.solve()

            opt.enable_vary(tag='stage1')
            opt.solve()

            opt.enable_vary(tag='stage2')
            opt.enable_targets(tag='stage2')
            opt.solve()
        else:
            opt.solve()

    return opt

def rematch_ir3(collider, line_name,
                boundary_conditions_left, boundary_conditions_right,
                mux_ir3, muy_ir3,
                alfx_ip3, alfy_ip3,
                betx_ip3, bety_ip3,
                dx_ip3, dpx_ip3,
                solve=True, staged_match=False, default_tol=None):

    assert line_name in ['lhcb1', 'lhcb2']
    bn = line_name[-2:]

    opt = collider[f'lhc{bn}'].match(
        solve=False,
        default_tol=default_tol,
        start=f's.ds.l3.{bn}', end=f'e.ds.r3.{bn}',
        init=boundary_conditions_left, init_at=xt.START,
        targets=[
            xt.TargetSet(at='ip3',
                    alfx=alfx_ip3, alfy=alfy_ip3, betx=betx_ip3, bety=bety_ip3,
                    dx=dx_ip3, dpx=dpx_ip3, tag='stage1'),
            xt.TargetSet(at=xt.END,
                    tars=('betx', 'bety', 'alfx', 'alfy', 'dx', 'dpx'),
                    value=boundary_conditions_right),
            xt.TargetRelPhaseAdvance('mux', mux_ir3),
            xt.TargetRelPhaseAdvance('muy', muy_ir3),
        ],
        vary=[
            xt.VaryList([f'kqt13.l3{bn}', f'kqt12.l3{bn}', f'kqtl11.l3{bn}',
                         f'kqtl10.l3{bn}', f'kqtl9.l3{bn}', f'kqtl8.l3{bn}',
                         f'kqtl7.l3{bn}', f'kq6.l3{bn}',
                         f'kq6.r3{bn}', f'kqtl7.r3{bn}',
                         f'kqtl8.r3{bn}', f'kqtl9.r3{bn}', f'kqtl10.r3{bn}',
                         f'kqtl11.r3{bn}', f'kqt12.r3{bn}', f'kqt13.r3{bn}'])
        ]
    )

    if solve:
        if staged_match:
            opt.disable_targets(tag='stage1')
            opt.solve()
            opt.enable_targets(tag='stage1')
            opt.solve()
        else:
            opt.solve()

    return opt

def rematch_ir4(collider, line_name,
                boundary_conditions_left, boundary_conditions_right,
                mux_ir4, muy_ir4,
                alfx_ip4, alfy_ip4,
                betx_ip4, bety_ip4,
                dx_ip4, dpx_ip4,
                solve=True, staged_match=False, default_tol=None):

    assert line_name in ['lhcb1', 'lhcb2']
    bn = line_name[-2:]

    opt = collider[f'lhc{bn}'].match(
        solve=False,
        default_tol=default_tol,
        start=f's.ds.l4.{bn}', end=f'e.ds.r4.{bn}',
        init=boundary_conditions_left, init_at=xt.START,
        targets=[
            xt.TargetSet(at='ip4',
                    alfx=alfx_ip4, alfy=alfy_ip4, betx=betx_ip4, bety=bety_ip4,
                    dx=dx_ip4, dpx=dpx_ip4, tag='stage1'),
            xt.TargetSet(at=xt.END,
                    tars=('betx', 'bety', 'alfx', 'alfy', 'dx', 'dpx'),
                    value=boundary_conditions_right),
            xt.TargetRelPhaseAdvance('mux', mux_ir4),
            xt.TargetRelPhaseAdvance('muy', muy_ir4),
        ],
        vary=[
            xt.VaryList([
                f'kqt13.l4{bn}', f'kqt12.l4{bn}', f'kqtl11.l4{bn}', f'kq10.l4{bn}',
                f'kq9.l4{bn}', f'kq8.l4{bn}', f'kq7.l4{bn}', f'kq6.l4{bn}',
                f'kq5.l4{bn}', f'kq5.r4{bn}', f'kq6.r4{bn}', f'kq7.r4{bn}',
                f'kq8.r4{bn}', f'kq9.r4{bn}', f'kq10.r4{bn}', f'kqtl11.r4{bn}',
                f'kqt12.r4{bn}', f'kqt13.r4{bn}'
            ])
        ]
    )

    if solve:
        if staged_match:
            opt.disable_targets(tag='stage1')
            opt.solve()
            opt.enable_targets(tag='stage1')
            opt.solve()
        else:
            opt.solve()

    return opt


def rematch_ir6(collider, line_name,
                boundary_conditions_left, boundary_conditions_right,
                mux_ir6, muy_ir6,
                alfx_ip6, alfy_ip6,
                betx_ip6, bety_ip6,
                dx_ip6, dpx_ip6,
                solve=True, staged_match=False, default_tol=None):

    assert line_name in ['lhcb1', 'lhcb2']
    bn = line_name[-2:]

    opt = collider[f'lhc{bn}'].match(
        solve=False,
        default_tol=default_tol,
        start=f's.ds.l6.{bn}', end=f'e.ds.r6.{bn}',
        # Left boundary
        init=boundary_conditions_left, init_at=xt.START,
        targets=[
            xt.TargetSet(at='ip6',
                    alfx=alfx_ip6, alfy=alfy_ip6, betx=betx_ip6, bety=bety_ip6,
                    dx=dx_ip6, dpx=dpx_ip6, tag='stage1'),
            xt.TargetSet(at=xt.END,
                    tars=('betx', 'bety', 'alfx', 'alfy', 'dx', 'dpx'),
                    value=boundary_conditions_right),
            xt.TargetRelPhaseAdvance('mux', mux_ir6),
            xt.TargetRelPhaseAdvance('muy', muy_ir6),
        ],
        vary=(
            [xt.VaryList([
                f'kqt13.l6{bn}', f'kqt12.l6{bn}', f'kqtl11.l6{bn}', f'kq10.l6{bn}',
                f'kq9.l6{bn}', f'kq8.l6{bn}', f'kq5.l6{bn}',
                f'kq5.r6{bn}', f'kq8.r6{bn}', f'kq9.r6{bn}',
                f'kq10.r6{bn}', f'kqtl11.r6{bn}', f'kqt12.r6{bn}', f'kqt13.r6{bn}'])]
            + [xt.Vary((f'kq4.r6{bn}' if bn == 'b1' else f'kq4.l6{bn}'))]
        )
    )

    # No staged match for IR6
    if solve:
        opt.solve()

    return opt

def rematch_ir7(collider, line_name,
              boundary_conditions_left, boundary_conditions_right,
              mux_ir7, muy_ir7,
              alfx_ip7, alfy_ip7,
              betx_ip7, bety_ip7,
              dx_ip7, dpx_ip7,
              solve=True, staged_match=False, default_tol=None):

    assert line_name in ['lhcb1', 'lhcb2']
    bn = line_name[-2:]

    opt = collider[f'lhc{bn}'].match(
        solve=False,
        default_tol=default_tol,
        start=f's.ds.l7.{bn}', end=f'e.ds.r7.{bn}',
        # Left boundary
        init=boundary_conditions_left, init_at=xt.START,
        targets=[
            xt.TargetSet(at='ip7',
                    alfx=alfx_ip7, alfy=alfy_ip7, betx=betx_ip7, bety=bety_ip7,
                    dx=dx_ip7, dpx=dpx_ip7, tag='stage1'),
            xt.TargetSet(at=xt.END,
                    tars=('betx', 'bety', 'alfx', 'alfy', 'dx', 'dpx'),
                    value=boundary_conditions_right),
            xt.TargetRelPhaseAdvance('mux', mux_ir7),
            xt.TargetRelPhaseAdvance('muy', muy_ir7),
        ],
        vary=[
            xt.VaryList([
                f'kqt13.l7{bn}', f'kqt12.l7{bn}', f'kqtl11.l7{bn}',
                f'kqtl10.l7{bn}', f'kqtl9.l7{bn}', f'kqtl8.l7{bn}',
                f'kqtl7.l7{bn}', f'kq6.l7{bn}',
                f'kq6.r7{bn}', f'kqtl7.r7{bn}',
                f'kqtl8.r7{bn}', f'kqtl9.r7{bn}', f'kqtl10.r7{bn}',
                f'kqtl11.r7{bn}', f'kqt12.r7{bn}', f'kqt13.r7{bn}'
            ])
        ]
    )

    if solve:
        if staged_match:
            opt.disable_targets(tag='stage1')
            opt.solve()
            opt.enable_targets(tag='stage1')
            opt.solve()
        else:
            opt.solve()

    return opt

def rematch_ir8(collider, line_name,
                boundary_conditions_left, boundary_conditions_right,
                mux_ir8, muy_ir8,
                alfx_ip8, alfy_ip8,
                betx_ip8, bety_ip8,
                dx_ip8, dpx_ip8,
                solve=True, staged_match=False, default_tol=None):

    assert line_name in ['lhcb1', 'lhcb2']
    bn = line_name[-2:]

    opt = collider[f'lhc{bn}'].match(
        solve=False,
        default_tol=default_tol,
        start=f's.ds.l8.{bn}', end=f'e.ds.r8.{bn}',
        # Left boundary
        init=boundary_conditions_left, init_at=xt.START,
        targets=[
            xt.TargetSet(at='ip8',
                    alfx=alfx_ip8, alfy=alfy_ip8, betx=betx_ip8, bety=bety_ip8,
                    dx=dx_ip8, dpx=dpx_ip8, tag='stage1'),
            xt.TargetSet(at=xt.END,
                    tars=('betx', 'bety', 'alfx', 'alfy', 'dx', 'dpx'),
                    value=boundary_conditions_right, tag='stage2'),
            xt.TargetRelPhaseAdvance('mux', mux_ir8),
            xt.TargetRelPhaseAdvance('muy', muy_ir8),
        ],
        vary=[
            xt.VaryList([
                f'kq6.l8{bn}', f'kq7.l8{bn}',
                f'kq8.l8{bn}', f'kq9.l8{bn}', f'kq10.l8{bn}', f'kqtl11.l8{bn}',
                f'kqt12.l8{bn}', f'kqt13.l8{bn}']
            ),
            xt.VaryList([f'kq4.l8{bn}', f'kq5.l8{bn}'], tag='stage1'),
            xt.VaryList([
                f'kq4.r8{bn}', f'kq5.r8{bn}', f'kq6.r8{bn}', f'kq7.r8{bn}',
                f'kq8.r8{bn}', f'kq9.r8{bn}', f'kq10.r8{bn}', f'kqtl11.r8{bn}',
                f'kqt12.r8{bn}', f'kqt13.r8{bn}'],
                tag='stage2')
        ]
    )

    if solve:
        if staged_match:
            opt.disable_targets(tag=['stage1', 'stage2'])
            opt.disable_vary(tag=['stage1', 'stage2'])
            opt.solve()

            opt.enable_vary(tag='stage1')
            opt.solve()

            opt.enable_targets(tag='stage1')
            opt.enable_targets(tag='stage2')
            opt.enable_vary(tag='stage2')
            opt.solve()
        else:
            opt.solve()

    return opt

def match_orbit_knobs_ip2_ip8(collider):

    all_knobs_ip2ip8 = ['acbxh3.r2', 'acbchs5.r2b1', 'pxip2b1', 'acbxh2.l8',
        'acbyhs4.r8b2', 'pyip2b1', 'acbxv1.l8', 'acbyvs4.l2b1', 'acbxh1.l8',
        'acbxv2.r8', 'pxip8b2', 'yip8b1', 'pxip2b2', 'acbcvs5.r2b1', 'acbyhs4.l8b1',
        'acbyvs4.l8b1', 'acbxh2.l2', 'acbxh3.l2', 'acbxv1.r8', 'acbxv1.r2',
        'acbyvs4.r2b2', 'acbyvs4.l2b2', 'yip8b2', 'xip2b2', 'acbxh2.r2',
        'acbyhs4.l2b2', 'acbxv2.r2', 'acbyhs5.r8b1', 'acbxh2.r8', 'acbxv3.r8',
        'acbyvs5.r8b2', 'acbyvs5.l2b2', 'yip2b1', 'acbxv2.l2', 'acbyhs4.r2b2',
        'acbyhs4.r2b1', 'xip8b2', 'acbyvs5.l2b1', 'acbyvs4.r8b1', 'acbyvs4.r8b2',
        'acbyvs5.r8b1', 'acbxh1.r8', 'acbyvs4.l8b2', 'acbyhs5.l2b1', 'acbyvs4.r2b1',
        'acbcvs5.r2b2', 'acbcvs5.l8b2', 'acbyhs4.r8b1', 'pxip8b1', 'acbxv1.l2',
        'yip2b2', 'acbyhs4.l8b2', 'acbxv3.r2', 'xip8b1', 'acbchs5.r2b2', 'acbxh3.l8',
        'acbxh3.r8', 'acbyhs5.r8b2', 'acbxv2.l8', 'acbxh1.l2', 'pyip8b1', 'pyip8b2',
        'acbxv3.l8', 'xip2b1', 'acbyhs5.l2b2', 'acbchs5.l8b2', 'acbcvs5.l8b1',
        'pyip2b2', 'acbxv3.l2', 'acbchs5.l8b1', 'acbyhs4.l2b1', 'acbxh1.r2']


    # kill all existing knobs
    for kk in all_knobs_ip2ip8:
        collider.vars[kk] = 0

    twinit_zero_orbit = [xt.TwissInit(), xt.TwissInit()]

    targets_close_bump = [
        xt.TargetSet(line='lhcb1', at=xt.END, x=0, px=0, y=0, py=0),
        xt.TargetSet(line='lhcb2', at=xt.END, x=0, px=0, y=0, py=0),
    ]

    bump_range_ip2 = {
        'start': ['s.ds.l2.b1', 's.ds.l2.b2'],
        'end': ['e.ds.r2.b1', 'e.ds.r2.b2'],
    }
    bump_range_ip8 = {
        'start': ['s.ds.l8.b1', 's.ds.l8.b2'],
        'end': ['e.ds.r8.b1', 'e.ds.r8.b2'],
    }

    correctors_ir2_single_beam_h = [
        'acbyhs4.l2b1', 'acbyhs4.r2b2', 'acbyhs4.l2b2', 'acbyhs4.r2b1',
        'acbyhs5.l2b2', 'acbyhs5.l2b1', 'acbchs5.r2b1', 'acbchs5.r2b2']

    correctors_ir2_single_beam_v = [
        'acbyvs4.l2b1', 'acbyvs4.r2b2', 'acbyvs4.l2b2', 'acbyvs4.r2b1',
        'acbyvs5.l2b2', 'acbyvs5.l2b1', 'acbcvs5.r2b1', 'acbcvs5.r2b2']

    correctors_ir8_single_beam_h = [
        'acbyhs4.l8b1', 'acbyhs4.r8b2', 'acbyhs4.l8b2', 'acbyhs4.r8b1',
        'acbchs5.l8b2', 'acbchs5.l8b1', 'acbyhs5.r8b1', 'acbyhs5.r8b2']

    correctors_ir8_single_beam_v = [
        'acbyvs4.l8b1', 'acbyvs4.r8b2', 'acbyvs4.l8b2', 'acbyvs4.r8b1',
        'acbcvs5.l8b2', 'acbcvs5.l8b1', 'acbyvs5.r8b1', 'acbyvs5.r8b2']

    correctors_ir2_common_h = [
        'acbxh1.l2', 'acbxh2.l2', 'acbxh3.l2', 'acbxh1.r2', 'acbxh2.r2', 'acbxh3.r2']

    correctors_ir2_common_v = [
        'acbxv1.l2', 'acbxv2.l2', 'acbxv3.l2', 'acbxv1.r2', 'acbxv2.r2', 'acbxv3.r2']

    correctors_ir8_common_h = [
        'acbxh1.l8', 'acbxh2.l8', 'acbxh3.l8', 'acbxh1.r8', 'acbxh2.r8', 'acbxh3.r8']

    correctors_ir8_common_v = [
        'acbxv1.l8', 'acbxv2.l8', 'acbxv3.l8', 'acbxv1.r8', 'acbxv2.r8', 'acbxv3.r8']

    #########################
    # Match IP offset knobs #
    #########################

    offset_match = 0.5e-3

    # ---------- on_o2v ----------

    opt_o2v = collider.match_knob(
        knob_name='on_o2v', knob_value_end=(offset_match * 1e3),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip2', y=offset_match, py=0),
            xt.TargetSet(line='lhcb2', at='ip2', y=offset_match, py=0),
        ]),
        vary=xt.VaryList(correctors_ir2_single_beam_v),
        run=False, init=twinit_zero_orbit, **bump_range_ip2,
    )
    opt_o2v.solve()
    opt_o2v.generate_knob()

    # ---------- on_o2h ----------

    opt_o2h = collider.match_knob(
        knob_name='on_o2h', knob_value_end=(offset_match * 1e3),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip2', x=offset_match, px=0),
            xt.TargetSet(line='lhcb2', at='ip2', x=offset_match, px=0),
        ]),
        vary=xt.VaryList(correctors_ir2_single_beam_h),
        run=False, init=twinit_zero_orbit, **bump_range_ip2,
    )
    opt_o2h.solve()
    opt_o2h.generate_knob()

    # ---------- on_o8v ----------

    opt_o8v = collider.match_knob(
        knob_name='on_o8v', knob_value_end=(offset_match * 1e3),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip8', y=offset_match, py=0),
            xt.TargetSet(line='lhcb2', at='ip8', y=offset_match, py=0),
        ]),
        vary=xt.VaryList(correctors_ir8_single_beam_v),
        run=False, init=twinit_zero_orbit, **bump_range_ip8,
    )
    opt_o8v.solve()
    opt_o8v.generate_knob()

    # ---------- on_o8h ----------

    opt_o8h = collider.match_knob(
        knob_name='on_o8h', knob_value_end=(offset_match * 1e3),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip8', x=offset_match, px=0),
            xt.TargetSet(line='lhcb2', at='ip8', x=offset_match, px=0),
        ]),
        vary=xt.VaryList(correctors_ir8_single_beam_h),
        run=False, init=twinit_zero_orbit, **bump_range_ip8,
    )
    opt_o8h.solve()
    opt_o8h.generate_knob()

    ##############################
    # Match angular offset knobs #
    ##############################

    ang_offset_match = 30e-6

    # ---------- on_a2h ----------

    opt_a2h = collider.match_knob(
        knob_name='on_a2h', knob_value_end=(ang_offset_match * 1e6),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip2', x=0, px=ang_offset_match),
            xt.TargetSet(line='lhcb2', at='ip2', x=0, px=ang_offset_match),
        ]),
        vary=xt.VaryList(correctors_ir2_single_beam_h),
        run=False, init=twinit_zero_orbit, **bump_range_ip2,
    )

    opt_a2h.solve()
    opt_a2h.generate_knob()

    # ---------- on_a2v ----------

    opt_a2v = collider.match_knob(
        knob_name='on_a2v', knob_value_end=(ang_offset_match * 1e6),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip2', y=0, py=ang_offset_match),
            xt.TargetSet(line='lhcb2', at='ip2', y=0, py=ang_offset_match),
        ]),
        vary=xt.VaryList(correctors_ir2_single_beam_v),
        run=False, init=twinit_zero_orbit, **bump_range_ip2,
    )

    opt_a2v.solve()
    opt_a2v.generate_knob()

    # ---------- on_a8h ----------

    opt_a8h = collider.match_knob(
        knob_name='on_a8h', knob_value_end=(ang_offset_match * 1e6),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip8', x=0, px=ang_offset_match),
            xt.TargetSet(line='lhcb2', at='ip8', x=0, px=ang_offset_match),
        ]),
        vary=xt.VaryList(correctors_ir8_single_beam_h),
        run=False, init=twinit_zero_orbit, **bump_range_ip8,
    )

    opt_a8h.solve()
    opt_a8h.generate_knob()

    # ---------- on_a8v ----------

    opt_a8v = collider.match_knob(
        knob_name='on_a8v', knob_value_end=(ang_offset_match * 1e6),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip8', y=0, py=ang_offset_match),
            xt.TargetSet(line='lhcb2', at='ip8', y=0, py=ang_offset_match),
        ]),
        vary=xt.VaryList(correctors_ir8_single_beam_v),
        run=False, init=twinit_zero_orbit, **bump_range_ip8,
    )

    opt_a8v.solve()
    opt_a8v.generate_knob()

    ##############################
    # Match crossing angle knobs #
    ##############################

    angle_match_ip2 = 170e-6
    angle_match_ip8 = 170e-6

    # ---------- on_x2h ----------

    opt_x2h = collider.match_knob(
        knob_name='on_x2h', knob_value_end=(angle_match_ip2 * 1e6),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip2',  x=0, px=angle_match_ip2),
            xt.TargetSet(line='lhcb2', at='ip2',  x=0, px=-angle_match_ip2),
        ]),
        vary=[
            xt.VaryList(correctors_ir2_single_beam_h),
            xt.VaryList(correctors_ir2_common_h, tag='mcbx')],
        run=False, init=twinit_zero_orbit, **bump_range_ip2,
    )
    # Set mcbx by hand
    testkqx2=abs(collider.varval['kqx.l2'])*7000./0.3
    acbx_xing_ir2 = 1.0e-6 if testkqx2 > 210. else 11.0e-6 # Value for 170 urad crossing
    for icorr in [1, 2, 3]:
        collider.vars[f'acbxh{icorr}.l2_from_on_x2h'] = acbx_xing_ir2
        collider.vars[f'acbxh{icorr}.r2_from_on_x2h'] = -acbx_xing_ir2
    # Match other correctors with fixed mcbx and generate knob
    opt_x2h.disable_vary(tag='mcbx')
    opt_x2h.solve()
    opt_x2h.generate_knob()

    # ---------- on_x2v ----------

    opt_x2v = collider.match_knob(
        knob_name='on_x2v', knob_value_end=(angle_match_ip2 * 1e6),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip2',  y=0, py=angle_match_ip2),
            xt.TargetSet(line='lhcb2', at='ip2',  y=0, py=-angle_match_ip2),
        ]),
        vary=[
            xt.VaryList(correctors_ir2_single_beam_v),
            xt.VaryList(correctors_ir2_common_v, tag='mcbx')],
        run=False, init=twinit_zero_orbit, **bump_range_ip2,
    )
    # Set mcbx by hand
    testkqx2=abs(collider.varval['kqx.l2'])*7000./0.3
    acbx_xing_ir2 = 1.0e-6 if testkqx2 > 210. else 11.0e-6
    for icorr in [1, 2, 3]:
        collider.vars[f'acbxv{icorr}.l2_from_on_x2v'] = acbx_xing_ir2
        collider.vars[f'acbxv{icorr}.r2_from_on_x2v'] = -acbx_xing_ir2
    # Match other correctors with fixed mcbx and generate knob
    opt_x2v.disable_vary(tag='mcbx')
    opt_x2v.solve()
    opt_x2v.generate_knob()

    # ---------- on_x8h ----------

    opt_x8h = collider.match_knob(
        knob_name='on_x8h', knob_value_end=(angle_match_ip8 * 1e6),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip8',  x=0, px=angle_match_ip8),
            xt.TargetSet(line='lhcb2', at='ip8',  x=0, px=-angle_match_ip8),
        ]),
        vary=[
            xt.VaryList(correctors_ir8_single_beam_h),
            xt.VaryList(correctors_ir8_common_h, tag='mcbx')],
        run=False, init=twinit_zero_orbit, **bump_range_ip8,
    )

    # Set mcbx by hand (reduce value by 10, to test matching algorithm)
    testkqx8=abs(collider.varval['kqx.l8'])*7000./0.3
    acbx_xing_ir8 = 1.0e-6 if testkqx8 > 210. else 11.0e-6 # Value for 170 urad crossing

    # Set mcbx by hand
    for icorr in [1, 2, 3]:
        collider.vars[f'acbxh{icorr}.l8_from_on_x8h'] = acbx_xing_ir8 * angle_match_ip8 / 170e-6
        collider.vars[f'acbxh{icorr}.r8_from_on_x8h'] = -acbx_xing_ir8 * angle_match_ip8 / 170e-6

    #   (reduce value by 10, to test matching algorithm)
    #   collider.vars[f'acbxh{icorr}.l8_from_on_x8h'] = acbx_xing_ir8 * angle_match_ip8 / 170e-6 * 0.1
    #   collider.vars[f'acbxh{icorr}.r8_from_on_x8h'] = -acbx_xing_ir8 * angle_match_ip8 / 170e-6 * 0.1

    # First round of optimization without changing mcbx
    opt_x8h.disable_vary(tag='mcbx')
    opt_x8h.step(3) # perform 3 steps without checking for convergence

    # Link all mcbx strengths to the first one
    collider.vars['acbxh2.l8_from_on_x8h'] =  collider.vars['acbxh1.l8_from_on_x8h']
    collider.vars['acbxh3.l8_from_on_x8h'] =  collider.vars['acbxh1.l8_from_on_x8h']
    collider.vars['acbxh2.r8_from_on_x8h'] = -collider.vars['acbxh1.l8_from_on_x8h']
    collider.vars['acbxh3.r8_from_on_x8h'] = -collider.vars['acbxh1.l8_from_on_x8h']
    collider.vars['acbxh1.r8_from_on_x8h'] = -collider.vars['acbxh1.l8_from_on_x8h']

    # Enable first mcbx knob (which controls the others)
    assert opt_x8h.vary[8].name == 'acbxh1.l8_from_on_x8h'
    opt_x8h.vary[8].active = True

    # Solve and generate knob
    opt_x8h.solve()
    opt_x8h.generate_knob()

    # ---------- on_x8v ----------

    opt_x8v = collider.match_knob(
        knob_name='on_x8v', knob_value_end=(angle_match_ip8 * 1e6),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip8', y=0, py=angle_match_ip8),
            xt.TargetSet(line='lhcb2', at='ip8', y=0, py=-angle_match_ip8),
        ]),
        vary=[
            xt.VaryList(correctors_ir8_single_beam_v),
            xt.VaryList(correctors_ir8_common_v, tag='mcbx')],
        run=False, init=twinit_zero_orbit, **bump_range_ip8,
    )

    # Set mcbx by hand
    testkqx8=abs(collider.varval['kqx.l8'])*7000./0.3
    acbx_xing_ir8 = 1.0e-6 if testkqx8 > 210. else 11.0e-6 # Value for 170 urad crossing
    # Set MCBX by hand
    for icorr in [1, 2, 3]:
        collider.vars[f'acbxv{icorr}.l8_from_on_x8v'] = acbx_xing_ir8 * angle_match_ip8 / 170e-6
        collider.vars[f'acbxv{icorr}.r8_from_on_x8v'] = -acbx_xing_ir8 * angle_match_ip8 / 170e-6

    # First round of optimization without changing mcbx
    opt_x8v.disable_vary(tag='mcbx')
    opt_x8v.step(3) # perform 3 steps without checking for convergence

    # Solve with all vary active and generate knob
    opt_x8v.enable_vary(tag='mcbx')
    opt_x8v.solve()
    opt_x8v.generate_knob()

    ##########################
    # Match separation knobs #
    ##########################

    sep_match = 2e-3

    # ---------- on_sep2h ----------

    opt_sep2h = collider.match_knob(
        knob_name='on_sep2h', knob_value_end=(sep_match * 1e3),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip2', x=sep_match, px=0),
            xt.TargetSet(line='lhcb2', at='ip2', x=-sep_match, px=0),
        ]),
        vary=[
            xt.VaryList(correctors_ir2_single_beam_h),
            xt.VaryList(correctors_ir2_common_h, tag='mcbx')],
        run=False, init=twinit_zero_orbit, **bump_range_ip2,
    )

    # Set mcbx by hand
    testkqx2=abs(collider.varval['kqx.l2'])*7000./0.3
    acbx_sep_ir2 = 18e-6 if testkqx2 > 210. else 16e-6

    for icorr in [1, 2, 3]:
        collider.vars[f'acbxh{icorr}.l2_from_on_sep2h'] = acbx_sep_ir2
        collider.vars[f'acbxh{icorr}.r2_from_on_sep2h'] = acbx_sep_ir2

    # Match other correctors with fixed mcbx and generate knob
    opt_sep2h.disable_vary(tag='mcbx')
    opt_sep2h.solve()
    opt_sep2h.generate_knob()

    # ---------- on_sep2v ----------

    opt_sep2v = collider.match_knob(
        knob_name='on_sep2v', knob_value_end=(sep_match * 1e3),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip2',  y=sep_match, py=0),
            xt.TargetSet(line='lhcb2', at='ip2',  y=-sep_match, py=0),
        ]),
        vary=[
            xt.VaryList(correctors_ir2_single_beam_v),
            xt.VaryList(correctors_ir2_common_v, tag='mcbx')],
        run=False, init=twinit_zero_orbit, **bump_range_ip2,
    )

    # Set mcbx by hand
    testkqx2=abs(collider.varval['kqx.l2'])*7000./0.3
    acbx_sep_ir2 = 18e-6 if testkqx2 > 210. else 16e-6

    for icorr in [1, 2, 3]:
        collider.vars[f'acbxv{icorr}.l2_from_on_sep2v'] = acbx_sep_ir2
        collider.vars[f'acbxv{icorr}.r2_from_on_sep2v'] = acbx_sep_ir2

    # Match other correctors with fixed mcbx and generate knob
    opt_sep2v.disable_vary(tag='mcbx')
    opt_sep2v.solve()
    opt_sep2v.generate_knob()

    # ---------- on_sep8h ----------

    opt_sep8h = collider.match_knob(
        knob_name='on_sep8h', knob_value_end=(sep_match * 1e3),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip8',  x=sep_match, px=0),
            xt.TargetSet(line='lhcb2', at='ip8',  x=-sep_match, px=0),
        ]),
        vary=[
            xt.VaryList(correctors_ir8_single_beam_h),
            xt.VaryList(correctors_ir8_common_h, tag='mcbx')],
        run=False, init=twinit_zero_orbit, **bump_range_ip8,
    )

    # Set mcbx by hand
    testkqx8=abs(collider.varval['kqx.l8'])*7000./0.3
    acbx_sep_ir8 = 18e-6 if testkqx8 > 210. else 16e-6

    for icorr in [1, 2, 3]:
        collider.vars[f'acbxh{icorr}.l8_from_on_sep8h'] = acbx_sep_ir8 * sep_match / 2e-3
        collider.vars[f'acbxh{icorr}.r8_from_on_sep8h'] = acbx_sep_ir8 * sep_match / 2e-3

    # Match other correctors with fixed mcbx and generate knob
    opt_sep8h.disable_vary(tag='mcbx')
    opt_sep8h.solve()
    opt_sep8h.generate_knob()

    # ---------- on_sep8v ----------

    opt_sep8v = collider.match_knob(
        knob_name='on_sep8v', knob_value_end=(sep_match * 1e3),
        targets=(targets_close_bump + [
            xt.TargetSet(line='lhcb1', at='ip8',  y=sep_match, py=0),
            xt.TargetSet(line='lhcb2', at='ip8',  y=-sep_match, py=0),
        ]),
        vary=[
            xt.VaryList(correctors_ir8_single_beam_v),
            xt.VaryList(correctors_ir8_common_v, tag='mcbx')],
        run=False, init=twinit_zero_orbit, **bump_range_ip8,
    )

    # Set mcbx by hand
    testkqx8=abs(collider.varval['kqx.l8'])*7000./0.3
    acbx_sep_ir8 = 18e-6 if testkqx8 > 210. else 16e-6

    for icorr in [1, 2, 3]:
        collider.vars[f'acbxv{icorr}.l8_from_on_sep8v'] = acbx_sep_ir8 * sep_match / 2e-3
        collider.vars[f'acbxv{icorr}.r8_from_on_sep8v'] = acbx_sep_ir8 * sep_match / 2e-3

    # Match other correctors with fixed mcbx and generate knob
    opt_sep8v.disable_vary(tag='mcbx')
    opt_sep8v.solve()
    opt_sep8v.generate_knob()

    v = collider.vars
    f = collider.functions
    v['phi_ir2'] = 90.
    v['phi_ir8'] = 0.
    for irn in [2, 8]:
        v[f'cphi_ir{irn}'] = f.cos(v[f'phi_ir{irn}'] * np.pi / 180.)
        v[f'sphi_ir{irn}'] = f.sin(v[f'phi_ir{irn}'] * np.pi / 180.)
        v[f'on_x{irn}h']   =  v[f'on_x{irn}'] * v[f'cphi_ir{irn}']
        v[f'on_x{irn}v']   =  v[f'on_x{irn}'] * v[f'sphi_ir{irn}']
        v[f'on_sep{irn}h'] = -v[f'on_sep{irn}'] * v[f'sphi_ir{irn}']
        v[f'on_sep{irn}v'] =  v[f'on_sep{irn}'] * v[f'cphi_ir{irn}']
        v[f'on_o{irn}h']   =  v[f'on_o{irn}'] * v[f'cphi_ir{irn}']
        v[f'on_o{irn}v']   =  v[f'on_o{irn}'] * v[f'sphi_ir{irn}']
        v[f'on_a{irn}h']   = -v[f'on_a{irn}'] * v[f'sphi_ir{irn}']
        v[f'on_a{irn}v']   =  v[f'on_a{irn}'] * v[f'cphi_ir{irn}']

    opt = {
        'on_o2h': opt_o2h, 'on_o2v': opt_o2v,
        'on_o8h': opt_o8h, 'on_o8v': opt_o8v,
        'on_a2h': opt_a2h, 'on_a2v': opt_a2v,
        'on_a8h': opt_a8h, 'on_a8v': opt_a8v,
        'on_x2h': opt_x2h, 'on_x2v': opt_x2v,
        'on_x8h': opt_x8h, 'on_x8v': opt_x8v,
        'on_sep2h': opt_sep2h, 'on_sep2v': opt_sep2v,
        'on_sep8h': opt_sep8h, 'on_sep8v': opt_sep8v,
    }

    return opt


import numpy as np

import xtrack as xt

import matplotlib.pyplot as plt

nemitt_x = 1e-6
nemitt_y = 1e-6


line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, p0c=7e12)
line.build_tracker()

# Some octupoles and chromaticity to see the footprint moving
line.vars['i_oct_b1'] = 500
line.match(
    targets=[xt.TargetList(['dqx', 'dqy'], value=10, tol=0.01)],
    vary=[xt.VaryList(['ksf.b1', 'ksd.b1'], step=1e-3)])

plt.close('all')
plt.figure(1)

# Compute and plot footprint on momentum
fp1 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                         freeze_longitudinal=True)
fp1.plot(color='b', label='delta=0')

# Compute and plot footprint off momentum
fp2 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                         freeze_longitudinal=True, delta0=1e-4)
fp2.plot(color='r', label='delta=1e-4')

assert np.allclose(line.record_last_track.delta[:], 1e-4, atol=1e-12, rtol=0)

# It is 12, 9 instead of 10, 10 because of non-linear chromaticity
assert np.isclose((np.max(fp2.qx) - np.max(fp1.qx))/1e-4, 12, atol=0.5, rtol=0)
assert np.isclose((np.max(fp2.qy) - np.max(fp1.qy))/1e-4, 9, atol=0.5, rtol=0)

plt.legend()

plt.show()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from enum import Enum
from pathlib import Path
from xobjects.general import _print  # noqa: F401

_pkg_root = Path(__file__).parent.absolute()


class _LOC:
    def __init__(self, name=None):
        self.name = name

    def __repr__(self):
        return self.name


START = _LOC('START')
END = _LOC('END')



import xtrack as xt
import xobjects as xo
import numpy as np
from scipy.constants import c as clight
from scipy.constants import e as qe

line = xt.Line.from_json('fccee_z_thick_with_sol_corrected.json')
n_turns_track_test = 6000
num_particles_test = 150

tt = line.get_table()
tt_bend = tt.rows[tt.element_type == 'RBend']
line.set(tt_bend, model='drift-kick-drift-expanded', integrator='uniform',
         num_multipole_kicks=3)

# line = xt.Line.from_json('fccee_t_with_sol_corrected.json')
# n_turns_track_test = 200
# num_particles_test = 200

tw_no_rad = line.twiss(method='4d')
line.configure_radiation(model='mean')
tt = line.get_table(attr=True)

line.vars['on_corr_ip.1'] = 1
line.vars['on_sol_ip.1'] = 1

# # Radiation only in solenoid
# ttmult = tt.rows[tt.element_type == 'Multipole']
# for nn in ttmult.name:
#     line[nn].radiation_flag=0

# RF on
line.vars['voltca1'] = line.vv['voltca1_ref']
line.vars['voltca2'] = line.vv['voltca2_ref']
line.compensate_radiation_energy_loss()
tw = line.twiss()

eloss = np.diff(tw.ptau) * line.particle_ref.energy0[0]
ds = tt.length[:-1]

mask_ds = ds > 0

dE_ds = eloss * 0
dE_ds[mask_ds] = -eloss[mask_ds] / ds[mask_ds]

tw_rad = line.twiss(eneloss_and_damping=True)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
ax1 = plt.subplot(3, 1, 1)
plt.plot(tw.s[:-1], dE_ds * 1e-2 * 1e-3, '.-', label='dE/ds')
plt.xlabel('s [m]')
plt.ylabel('dE/ds [keV/m]')
ax2 = plt.subplot(3, 1, 2, sharex=ax1)
plt.plot(tw.s, tt.ks, '.-', label='ks')
plt.xlabel('s [m]')
ax3 = plt.subplot(3, 1, 3, sharex=ax1)
plt.plot(tw.s, tw_rad.delta, '.-', label='ks')
plt.xlabel('s [m]')

print('partition numbers: ', tw_rad.partition_numbers)
print('gemit_x: ', tw_rad.eq_gemitt_x)
print('gemit_y: ', tw_rad.eq_gemitt_y)

tw_rad = line.twiss(eneloss_and_damping=True, radiation_method='full')

ex = tw_rad.eq_gemitt_x
ey = tw_rad.eq_gemitt_y
ez = tw_rad.eq_gemitt_zeta

tw_integ = line.twiss(radiation_integrals = True)

# xo.assert_allclose(tw_integ.rad_int_eq_gemitt_x, ex, rtol=5e-2, atol=0)
# xo.assert_allclose(tw_integ.rad_int_eq_gemitt_y, ey, rtol=5e-2, atol=0)

# Equilibrium beam sizes
beam_sizes = tw_rad.get_beam_covariance(
    gemitt_x=tw_rad.eq_gemitt_x, gemitt_y=tw_rad.eq_gemitt_y,
    gemitt_zeta=tw_rad.eq_gemitt_zeta)

line.configure_radiation(model='quantum')
p = line.build_particles(num_particles=num_particles_test)

# Switch to multithreaded
line.discard_tracker()
line.build_tracker(_context=xo.ContextCpu(omp_num_threads='auto'),
                   use_prebuilt_kernels=False)

line.track(p, num_turns=n_turns_track_test, turn_by_turn_monitor=True, time=True,
           with_progress=10)
mon_at_start = line.record_last_track
print(f'Tracking time: {line.time_last_track}')

# Approximated calculation based on Hofmann Eq. 14.19
twe = tw_no_rad
tt = line.get_table(attr=True)
hh = np.sqrt(np.diff(twe.px, append=0)**2 + np.diff(twe.py, append=0)**2)
dl = np.diff(twe.s, append=line.get_length())
gamma0 = line.particle_ref.gamma0[0]

# dyprime = twe.dpy*(1 - twe.delta) - twe.py # This is wrong!!!!!
dyprime = twe.dpy*(1 - twe.delta) - twe.kin_py

cur_H_y = twe.gamy * twe.dy**2 + 2 * twe.alfy * twe.dy * dyprime + twe.bety * dyprime**2
I5_y  = np.sum(cur_H_y * hh**3 * dl)
I2_y = np.sum(hh**2 * dl)
I4_y = np.sum(twe.dy * hh**3 * dl) # to be generalized for combined function magnets

lam_comp = 2.436e-12 # [m]
ey_hof = 55 * np.sqrt(3) / 96 * lam_comp / 2 / np.pi * gamma0**2 * I5_y / (I2_y - I4_y)

beam_sizes_hof = tw_rad.get_beam_covariance(
    gemitt_x=tw_rad.eq_gemitt_x, gemitt_y=ey_hof,
    gemitt_zeta=tw_rad.eq_gemitt_zeta)

# Plots
import matplotlib.pyplot as plt
plt.close('all')

line.configure_radiation(model='mean') # Leave the line in a twissable state
mon = line.record_last_track

fig = plt.figure(figsize=(6.4, 4.8*1.3))

spx = fig. add_subplot(3, 1, 1)
spx.plot(1e6 * np.std(mon.x, axis=0), label='track')
spx.axhline(1e6 * beam_sizes['sigma_x', 'ip.1'], color='red', label='twiss')
spx.legend(loc='lower right', fontsize='small')
spx.set_ylabel(r'$\sigma_{x}$ [$\mu m$]')
spx.set_ylim(bottom=0)

spy = fig. add_subplot(3, 1, 2, sharex=spx)
spy.plot(1e9 * np.std(mon.y, axis=0), label='track')
spy.axhline(1e9 * beam_sizes['sigma_y', 'ip.1'], color='red', label='twiss')
# spy.axhline(1e9 * beam_sizes_hof['sigma_y', 'ip.1'], color='red', label='twiss')
spy.set_ylabel(r'$\sigma_{y}$ [nm]')
spy.set_ylim(bottom=0)

spz = fig. add_subplot(3, 1, 3, sharex=spx)
spz.plot(np.std(1e3*mon.zeta, axis=0))
spz.axhline(1e3*beam_sizes['sigma_zeta', 'ip.1'], color='red', label='twiss')
spz.set_ylabel(r'$\sigma_{z}$ [mm]')
spz.set_ylim(bottom=0)
spz.set_xlabel('s [m]')
plt.subplots_adjust(left=.2, top=.95, hspace=.2)


plt.show()


import numpy as np
from cpymad.madx import Madx
import xtrack as xt

mad_ti2 = Madx()
mad_ti2.call('../../test_data/sps_to_lhc_ti2/ti2.seq')
mad_ti2.call('../../test_data/sps_to_lhc_ti2/ti2_liu.str')
mad_ti2.beam()
mad_ti2.use('ti2')

line = xt.Line.from_madx_sequence(mad_ti2.sequence['ti2'])
line.particle_ref = xt.Particles(p0c=450e9, mass0=xt.PROTON_MASS_EV, q0=1)
tt = line.get_table()

# Define elements to be used as monitors for orbit correction
# (in this case all element names starting by "bpm" and not ending by "_entry" or "_exit")
tt_monitors = tt.rows['bpm.*'].rows['.*(?<!_entry)$'].rows['.*(?<!_exit)$']
line.steering_monitors_x = tt_monitors.name
line.steering_monitors_y = tt_monitors.name

# Define elements to be used as correctors for orbit correction
# (in this case all element names starting by "mci.", containing "h." or "v.")
tt_h_correctors = tt.rows['mci.*'].rows['.*h\..*']
line.steering_correctors_x = tt_h_correctors.name
tt_v_correctors = tt.rows['mci.*'].rows['.*v\..*']
line.steering_correctors_y = tt_v_correctors.name

# Initial conditions from upstream ring
init = xt.TwissInit(betx=27.77906807, bety=120.39920690,
                     alfx=0.63611880, alfy=-2.70621900,
                     dx=-0.59866300, dpx=0.01603536)

# Reference twiss (no misalignments)
tw_ref = line.twiss4d(start='ti2$start', end='ti2$end', init=init)

# Introduce misalignments on all quadrupoles
tt = line.get_table()
tt_quad = tt.rows['mqi.*']
rgen = np.random.RandomState(2) # fix seed for random number generator
shift_x = rgen.randn(len(tt_quad)) * 0.1e-3 # 0.1 mm rms shift on all quads
shift_y = rgen.randn(len(tt_quad)) * 0.1e-3 # 0.1 mm rms shift on all quads
for nn_quad, sx, sy in zip(tt_quad.name, shift_x, shift_y):
    line[nn_quad].shift_x = sx
    line[nn_quad].shift_y = sy

# Twiss before correction
tw_before = line.twiss4d(start='ti2$start', end='ti2$end', init=init)

# Correct trajectory
correction = line.correct_trajectory(twiss_table=tw_ref, start='ti2$start', end='ti2$end')
# prints:
#
# Iteration 0, x_rms: 2.01e-03 -> 1.80e-04, y_rms: 6.94e-04 -> 1.23e-04
# Iteration 1, x_rms: 1.80e-04 -> 1.80e-04, y_rms: 1.23e-04 -> 1.23e-04

# Twiss after correction
tw_after = line.twiss4d(start='ti2$start', end='ti2$end', init=init)

# Extract correction strength
s_x_correctors = correction.x_correction.s_correctors
s_y_correctors = correction.y_correction.s_correctors
kicks_x = correction.x_correction.get_kick_values()
kicks_y = correction.y_correction.get_kick_values()

#!end-doc-part

# Plots
import matplotlib.pyplot as plt
plt.close('all')

plt.figure(1, figsize=(6.4, 4.8*1.7))
sp1 = plt.subplot(411)
sp1.plot(tw_before.s, tw_before.x * 1e3, label='before corr.')
sp1.plot(tw_after.s, tw_after.x * 1e3, label='after corr.')
plt.legend(loc='upper left')
plt.ylabel('x [mm]')

sp2 = plt.subplot(412, sharex=sp1)
sp2.stem(s_x_correctors, kicks_x * 1e6)
plt.ylabel(r'x kick [$\mu$rad]')

sp3 = plt.subplot(413, sharex=sp1, sharey=sp1)
sp3.plot(tw_before.s, tw_before.y * 1e3)
sp3.plot(tw_after.s, tw_after.y * 1e3)
plt.ylabel('y [mm]')

sp4 = plt.subplot(414, sharex=sp1)
sp4.stem(s_y_correctors, kicks_y * 1e6)
plt.ylabel(r'y kick [$\mu$rad]')
sp4.set_xlabel('s [m]')

plt.subplots_adjust(hspace=0.3, top=0.95, bottom=0.08)

plt.show()



def set_var_limits_and_steps(collider):

    nrj = 7000.
    scale = 23348.89927
    scmin = 0.03*7000./nrj
    qtlimitx28 = 1.0*225.0/scale
    qtlimitx15 = 1.0*205.0/scale
    qtlimit2 = 1.0*160.0/scale
    qtlimit3 = 1.0*200.0/scale
    qtlimit4 = 1.0*125.0/scale
    qtlimit5 = 1.0*120.0/scale
    qtlimit6 = 1.0*90.0/scale

    limitmcbxh =  63.5988e-6
    limitmcbxv =  67.0164e-6
    limitmcbx  =  67.0164e-6
    limitmcby  =  96.3000e-6
    limitmcb   =  80.8000e-6
    limitmcbc  =  89.8700e-6
    limitmcbw  =  80.1400e-6

    collider.vars.vary_default.update({
        'kqtf.a12b1': {'step': 1e-8, 'limits': None},
        'kqtf.a12b2': {'step': 1e-8, 'limits': None},
        'kqtd.a12b1': {'step': 1e-8, 'limits': None},
        'kqtd.a12b2': {'step': 1e-8, 'limits': None},
        'kqtf.a23b1': {'step': 1e-8, 'limits': None},
        'kqtf.a23b2': {'step': 1e-8, 'limits': None},
        'kqtd.a23b1': {'step': 1e-8, 'limits': None},
        'kqtd.a23b2': {'step': 1e-8, 'limits': None},
        'kqtf.a34b1': {'step': 1e-8, 'limits': None},
        'kqtf.a34b2': {'step': 1e-8, 'limits': None},
        'kqtd.a34b1': {'step': 1e-8, 'limits': None},
        'kqtd.a34b2': {'step': 1e-8, 'limits': None},
        'kqtf.a45b1': {'step': 1e-8, 'limits': None},
        'kqtf.a45b2': {'step': 1e-8, 'limits': None},
        'kqtd.a45b1': {'step': 1e-8, 'limits': None},
        'kqtd.a45b2': {'step': 1e-8, 'limits': None},
        'kqtf.a56b1': {'step': 1e-8, 'limits': None},
        'kqtf.a56b2': {'step': 1e-8, 'limits': None},
        'kqtd.a56b1': {'step': 1e-8, 'limits': None},
        'kqtd.a56b2': {'step': 1e-8, 'limits': None},
        'kqtf.a67b1': {'step': 1e-8, 'limits': None},
        'kqtf.a67b2': {'step': 1e-8, 'limits': None},
        'kqtd.a67b1': {'step': 1e-8, 'limits': None},
        'kqtd.a67b2': {'step': 1e-8, 'limits': None},
        'kqtf.a78b1': {'step': 1e-8, 'limits': None},
        'kqtf.a78b2': {'step': 1e-8, 'limits': None},
        'kqtd.a78b1': {'step': 1e-8, 'limits': None},
        'kqtd.a78b2': {'step': 1e-8, 'limits': None},
        'kqtf.a81b1': {'step': 1e-8, 'limits': None},
        'kqtf.a81b2': {'step': 1e-8, 'limits': None},
        'kqtd.a81b1': {'step': 1e-8, 'limits': None},
        'kqtd.a81b2': {'step': 1e-8, 'limits': None},
        'kqf.a12': {'step': 1e-10, 'limits': None},
        'kqd.a12': {'step': 1e-10, 'limits': None},
        'kqf.a23': {'step': 1e-10, 'limits': None},
        'kqd.a23': {'step': 1e-10, 'limits': None},
        'kqf.a34': {'step': 1e-10, 'limits': None},
        'kqd.a34': {'step': 1e-10, 'limits': None},
        'kqf.a45': {'step': 1e-10, 'limits': None},
        'kqd.a45': {'step': 1e-10, 'limits': None},
        'kqf.a56': {'step': 1e-10, 'limits': None},
        'kqd.a56': {'step': 1e-10, 'limits': None},
        'kqf.a67': {'step': 1e-10, 'limits': None},
        'kqd.a67': {'step': 1e-10, 'limits': None},
        'kqf.a78': {'step': 1e-10, 'limits': None},
        'kqd.a78': {'step': 1e-10, 'limits': None},
        'kqf.a81': {'step': 1e-10, 'limits': None},
        'kqd.a81': {'step': 1e-10, 'limits': None},

        'kq5.l2b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq4.l2b1':    {'step': 1.0E-6, 'limits': (-qtlimit2,       -qtlimit2*scmin)},
        'kq6.l2b1':    {'step': 1.0E-6, 'limits': (-qtlimit2,       -qtlimit2*scmin)},
        'kq7.l2b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq8.l2b1':    {'step': 1.0E-6, 'limits': (-qtlimit3,       -qtlimit3*scmin*0.9)},
        'kq9.l2b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq10.l2b1':    {'step': 1.0E-6, 'limits': (-qtlimit3,       -qtlimit3*scmin)},
        'kqtl11.l2b1': {'step': 1.0E-6, 'limits': (-qtlimit4,       qtlimit4)},
        'kqt12.l2b1':  {'step': 1.0E-6, 'limits': (-qtlimit5,       qtlimit5)},
        'kqt13.l2b1':  {'step': 1.0E-6, 'limits': (-qtlimit5,       qtlimit5)},
        'kq4.r2b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq5.r2b1':    {'step': 1.0E-6, 'limits': (-qtlimit2,       -qtlimit2*scmin)},
        'kq6.r2b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq7.r2b1':    {'step': 1.0E-6, 'limits': (-qtlimit3,       -qtlimit3*scmin)},
        'kq8.r2b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq9.r2b1':    {'step': 1.0E-6, 'limits': (-qtlimit3,       -qtlimit3*scmin)},
        'kq10.r2b1':   {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kqtl11.r2b1': {'step': 1.0E-6, 'limits': (-qtlimit4,       qtlimit4)},
        'kqt12.r2b1':  {'step': 1.0E-6, 'limits': (-qtlimit5,       qtlimit5)},
        'kqt13.r2b1':  {'step': 1.0E-6, 'limits': (-qtlimit5,       qtlimit5)},
        'kqt13.l2b2':  {'step': 1.0E-6, 'limits': (-qtlimit5,       qtlimit5)},
        'kqt12.l2b2':  {'step': 1.0E-6, 'limits': (-qtlimit5,       qtlimit5)},
        'kqtl11.l2b2': {'step': 1.0E-6, 'limits': (-qtlimit4,       qtlimit4)},
        'kq10.l2b2':   {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq9.l2b2':    {'step': 1.0E-6, 'limits': (-qtlimit3,       -qtlimit3*scmin)},
        'kq8.l2b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq7.l2b2':    {'step': 1.0E-6, 'limits': (-qtlimit3,       -qtlimit3*scmin)},
        'kq6.l2b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq5.l2b2':    {'step': 1.0E-6, 'limits': (-qtlimit2,       -qtlimit2*scmin)},
        'kq4.l2b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq4.r2b2':    {'step': 1.0E-6, 'limits': (-qtlimit2,       -qtlimit2*scmin)},
        'kq5.r2b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq6.r2b2':    {'step': 1.0E-6, 'limits': (-qtlimit2,       -qtlimit2*scmin)},
        'kq7.r2b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq8.r2b2':    {'step': 1.0E-6, 'limits': (-qtlimit3,       -qtlimit3*scmin)},
        'kq9.r2b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq10.r2b2':   {'step': 1.0E-6, 'limits': (-qtlimit3,       -qtlimit3*scmin)},
        'kqtl11.r2b2': {'step': 1.0E-6, 'limits': (-qtlimit4,       qtlimit4)},
        'kqt12.r2b2':  {'step': 1.0E-6, 'limits': (-qtlimit5,       qtlimit5)},
        'kqt13.r2b2':  {'step': 1.0E-6, 'limits': (-qtlimit5,       qtlimit5)},

        'kqt13.l3b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqt12.l3b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqtl11.l3b1': {'step': 1.0E-9, 'limits': (-qtlimit4*400./550., qtlimit4*400./550.)},
        'kqtl10.l3b1': {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl9.l3b1':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl8.l3b1':  {'step': 1.0E-9, 'limits': (-qtlimit4*450./550., qtlimit4*450./550.)},
        'kqtl7.l3b1':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kq6.l3b1':    {'step': 1.0E-9, 'limits': (-qtlimit6, qtlimit6)},
        'kq6.r3b1':    {'step': 1.0E-9, 'limits': (-qtlimit6, qtlimit6)},
        'kqtl7.r3b1':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl8.r3b1':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl9.r3b1':  {'step': 1.0E-9, 'limits': (-qtlimit4*500./550., qtlimit4*500./550.)},
        'kqtl10.r3b1': {'step': 1.0E-9, 'limits': (-qtlimit4*450./550., qtlimit4*450./550.)},
        'kqtl11.r3b1': {'step': 1.0E-9, 'limits': (-qtlimit4*520./550., qtlimit4*520./550.)},
        'kqt12.r3b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r3b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.l3b2':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqt12.l3b2':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqtl11.l3b2': {'step': 1.0E-9, 'limits': (-qtlimit4*400./550., qtlimit4*400./550.)},
        'kqtl10.l3b2': {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl9.l3b2':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl8.l3b2':  {'step': 1.0E-9, 'limits': (-qtlimit4*450./550., qtlimit4*450./550.)},
        'kqtl7.l3b2':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kq6.l3b2':    {'step': 1.0E-9, 'limits': (-qtlimit6, qtlimit6)},
        'kq6.r3b2':    {'step': 1.0E-9, 'limits': (-qtlimit6, qtlimit6)},
        'kqtl7.r3b2':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl8.r3b2':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl9.r3b2':  {'step': 1.0E-9, 'limits': (-qtlimit4*500./550., qtlimit4*500./550.)},
        'kqtl10.r3b2': {'step': 1.0E-9, 'limits': (-qtlimit4*450./550., qtlimit4*450./550.)},
        'kqtl11.r3b2': {'step': 1.0E-9, 'limits': (-qtlimit4*520./550., qtlimit4*520./550.)},
        'kqt12.r3b2':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r3b2':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},

        'kqt13.l4b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt12.l4b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqtl11.l4b1': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kq10.l4b1':   {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq9.l4b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq8.l4b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq7.l4b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq6.l4b1':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq5.l4b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq5.r4b1':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq6.r4b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq7.r4b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq8.r4b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq9.r4b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq10.r4b1':   {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kqtl11.r4b1': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.r4b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r4b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.l4b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt12.l4b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqtl11.l4b2': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kq10.l4b2':   {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq9.l4b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq8.l4b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq7.l4b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq6.l4b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq5.l4b2':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq5.r4b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq6.r4b2':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq7.r4b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq8.r4b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq9.r4b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq10.r4b2':   {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kqtl11.r4b2': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.r4b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r4b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},

        'kqt13.l6b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt12.l6b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqtl11.l6b1': {'step': 1.0E-6, 'limits': (-qtlimit4*400./550., qtlimit4*400./550.)},
        'kq10.l6b1':   {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq9.l6b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq8.l6b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq5.l6b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2*1.05)}, # 5% more (used in the nominal optics...)
        'kq4.l6b1':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq4.r6b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq5.r6b1':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq8.r6b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq9.r6b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq10.r6b1':   {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kqtl11.r6b1': {'step': 1.0E-6, 'limits': (-qtlimit4*300./550., qtlimit4*300./550.)},
        'kqt12.r6b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r6b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.l6b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt12.l6b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqtl11.l6b2': {'step': 1.0E-6, 'limits': (-qtlimit4*400./550., qtlimit4*400./550.)},
        'kq10.l6b2':   {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq9.l6b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq8.l6b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq5.l6b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq4.l6b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq4.r6b2':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq5.r6b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq8.r6b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq9.r6b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq10.r6b2':   {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kqtl11.r6b2': {'step': 1.0E-6, 'limits': (-qtlimit4*300./550., qtlimit4*300./550.)},
        'kqt12.r6b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r6b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},

        'kqt13.l7b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqt12.l7b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqtl11.l7b1': {'step': 1.0E-9, 'limits': (-qtlimit4*300./550., qtlimit4*300./550.)},
        'kqtl10.l7b1': {'step': 1.0E-9, 'limits': (-qtlimit4*500./550., qtlimit4*500./550.)},
        'kqtl9.l7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4*400./550., qtlimit4*400./550.)},
        'kqtl8.l7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4*300./550., qtlimit4*300./550.)},
        'kqtl7.l7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kq6.l7b1':    {'step': 1.0E-9, 'limits': (-qtlimit6, qtlimit6)},
        'kq6.r7b1':    {'step': 1.0E-9, 'limits': (-qtlimit6, qtlimit6)},
        'kqtl7.r7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl8.r7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4*550./550., qtlimit4*550./550.)},
        'kqtl9.r7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4*500./550., qtlimit4*500./550.)},
        'kqtl10.r7b1': {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl11.r7b1': {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.r7b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r7b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.l7b2':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqt12.l7b2':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqtl11.l7b2': {'step': 1.0E-9, 'limits': (-qtlimit4*300./550., qtlimit4*300./550.)},
        'kqtl10.l7b2': {'step': 1.0E-9, 'limits': (-qtlimit4*500./550., qtlimit4*500./550.)},
        'kqtl9.l7b2':  {'step': 1.0E-9, 'limits': (-qtlimit4*400./550., qtlimit4*400./550.)},
        'kqtl8.l7b2':  {'step': 1.0E-9, 'limits': (-qtlimit4*300./550., qtlimit4*300./550.)},
        'kqtl7.l7b2':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kq6.l7b2':    {'step': 1.0E-9, 'limits': (-qtlimit6, qtlimit6)},
        'kq6.r7b2':    {'step': 1.0E-9, 'limits': (-qtlimit6, qtlimit6)},
        'kqtl7.r7b2':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl8.r7b2':  {'step': 1.0E-9, 'limits': (-qtlimit4*550./550., qtlimit4*550./550.)},
        'kqtl9.r7b2':  {'step': 1.0E-9, 'limits': (-qtlimit4*500./550., qtlimit4*500./550.)},
        'kqtl10.r7b2': {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqtl11.r7b2': {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.r7b2':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r7b2':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},

        'kq4.r8b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq5.r8b1':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq6.r8b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq7.r8b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq8.r8b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq9.r8b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq10.r8b1':   {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kqtl11.r8b1': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.r8b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r8b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kq4.l8b1':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq5.l8b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq6.l8b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq7.l8b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq8.l8b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq9.l8b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq10.l8b1':   {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kqtl11.l8b1': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.l8b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.l8b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kq4.r8b2':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq5.r8b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq6.r8b2':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq7.r8b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq8.r8b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq9.r8b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq10.r8b2':   {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kqtl11.r8b2': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.r8b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r8b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kq5.l8b2':    {'step': 1.0E-6, 'limits': (-qtlimit2, qtlimit2*scmin)},
        'kq4.l8b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq6.l8b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq7.l8b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq8.l8b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq9.l8b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, qtlimit3*scmin)},
        'kq10.l8b2':   {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kqtl11.l8b2': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.l8b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.l8b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},

        # Single beam correctors at IR2 (Q4 and Q5, H and V)
        'acbyvs4.l2b1':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbyvs4.r2b2':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbyvs4.l2b2':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbyvs4.r2b1':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbyvs5.l2b2':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbyvs5.l2b1':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbcvs5.r2b1':{'step': 1.0e-15, 'limits': (-limitmcbc, limitmcbc)},
        'acbcvs5.r2b2':{'step': 1.0e-15, 'limits': (-limitmcbc, limitmcbc)},
        'acbyhs4.l8b1':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbyhs4.r8b2':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbyhs4.l8b2':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbyhs4.r8b1':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbchs5.l8b2':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbchs5.l8b1':{'step': 1.0e-15, 'limits': (-limitmcby, limitmcby)},
        'acbyhs5.r8b1':{'step': 1.0e-15, 'limits': (-limitmcbc, limitmcbc)},
        'acbyhs5.r8b2':{'step': 1.0e-15, 'limits': (-limitmcbc, limitmcbc)},

        # Correctors in IR8 (Q4 and Q5, and triplet)
        'acbyhs4.l8b1':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbyhs4.r8b2':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbyhs4.l8b2':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbyhs4.r8b1':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbchs5.l8b2':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbchs5.l8b1':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbyhs5.r8b1':{'step': 1.0e-15,'limits':(-limitmcbc, limitmcbc)},
        'acbyhs5.r8b2':{'step': 1.0e-15,'limits':(-limitmcbc, limitmcbc)},
        'acbxh1.l8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxh2.l8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxh3.l8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxh1.r8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxh2.r8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxh3.r8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbyvs4.l8b1':{'step': 1.0e-15, 'limits':(-limitmcby, limitmcby)},
        'acbcvs5.l8b1':{'step': 1.0e-15, 'limits':(-limitmcbc, limitmcbc)},
        'acbyvs4.r8b1':{'step': 1.0e-15, 'limits':(-limitmcby, limitmcby)},
        'acbyvs5.r8b1':{'step': 1.0e-15, 'limits':(-limitmcby, limitmcby)},
        'acbyvs4.l8b2':{'step': 1.0e-15, 'limits':(-limitmcby, limitmcby)},
        'acbcvs5.l8b2':{'step': 1.0e-15, 'limits':(-limitmcbc, limitmcbc)},
        'acbyvs4.r8b2':{'step': 1.0e-15, 'limits':(-limitmcby, limitmcby)},
        'acbyvs5.r8b2':{'step': 1.0e-15, 'limits':(-limitmcby, limitmcby)},
        'acbxv1.l8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxv2.l8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxv3.l8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxv1.r8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxv2.r8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxv3.r8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
    })



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt

# Load collider with two lines
collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider.build_trackers()
collider.lhcb1.twiss_default.clear() # clear twiss default settings
collider.lhcb2.twiss_default.clear() # clear twiss default settings
collider.vars['on_disp'] = 0 # disable dispersion correction

# Set a horizontal crossing angle between the two beams  in the interaction
# point ip1 and a vertical crossing angle in ip5
collider.vars['on_x1hl'] = 10 # [urad]
collider.vars['on_x5vl'] = 10 # [urad]
# Set a vertical separation between the two beams in the interaction point ip5
# and a horizontal separation in ip5
collider.vars['on_sep1v'] = 0.5 # [mm]
collider.vars['on_sep5h'] = 0.5 # [mm]

# Twiss the two lines (4d method since cavities are off)
tw1 = collider.lhcb1.twiss(method='4d')
tw2 = collider.lhcb2.twiss(method='4d')

tw1.reference_frame # is `proper`
tw2.reference_frame # is `proper`

# tw1 has a clokwise orientation while tw2 has a counter-clockwise orientation.
#
# name         s   mux   muy         x          y         px         py
# ip1          0     0     0 4.313e-09     0.0005  1.002e-05 -4.133e-09
# ip3       6665 15.95 15.45 2.392e-08 -2.209e-07  3.695e-10 -3.005e-09
# ip5  1.333e+04 30.93 29.99    0.0005  4.332e-09  1.918e-08      1e-05
# ip7  1.999e+04 46.35 44.59 2.138e-07 -1.785e-08 -2.132e-09 -3.078e-11

tw2.rows['ip[1,3,5,7]'].cols['s mux muy x y px py'].show(digits=4)
# prints:
#
# name         s   mux   muy          x          y         px         py
# ip7       6665 16.19 15.45 -5.941e-09   2.69e-07 -1.098e-10  2.428e-09
# ip5  1.333e+04 31.28 30.37     0.0005  5.997e-09 -4.562e-09  1.003e-05
# ip3  1.999e+04 46.46 44.81  -9.85e-08 -2.447e-08  3.218e-09 -9.711e-10
# ip1  2.666e+04 62.31 60.32 -2.278e-09    -0.0005     -1e-05   2.57e-08

# -- Reverse b2 twiss --
# The `reverse`` flag, allows getting the output of the twiss in the counter-rotating
# reference system. When `reverse` is True, the ordering of the elements is reversed,
# the zero of the s coordinate and fo the phase advances is set at  the new start,
# the sign of the coordinates s and x is inverted, while the sign of the coordinate
# y is unchanged. In symbols:
#
# s   -> L - s
# mux -> mux(L) - mux
# muy -> muy(L) - muy
# x   -> -x
# y   -> y
# px  -> px
# py  -> -py

tw2_r = collider.lhcb2.twiss(method='4d', reverse=True)
tw2_r.reference_frame # is `reverse`

tw2_r.rows['ip[1,3,5,7]'].cols['s mux muy x y px py'].show(digits=4)
# prints:
#
# name         s   mux   muy         x          y         px         py
# ip1  2.183e-11     0     0 2.278e-09    -0.0005     -1e-05  -2.57e-08
# ip3       6665 15.85 15.51  9.85e-08 -2.447e-08  3.218e-09  9.711e-10
# ip5  1.333e+04 31.03 29.95   -0.0005  5.997e-09 -4.562e-09 -1.003e-05
# ip7  1.999e+04 46.12 44.87 5.941e-09   2.69e-07 -1.098e-10 -2.428e-09

# In this way, for a collider, it is possible to plot the closed orbit and the
# twiss functions of the two beams in the same graph. For example:

import matplotlib.pyplot as plt
plt.close('all')
fig = plt.figure(1)

ax1 = fig.add_subplot(211)
ax1.plot(tw1.s, tw1.x * 1000, color='b', label='b1')
ax1.plot(tw2_r.s, tw2_r.x * 1000, color='r', label='b2')
ax1.set_ylabel('x [m]')
ax1.legend(loc='best')

ax2 = fig.add_subplot(212, sharex=ax1)
ax2.plot(tw1.s, tw1.y * 1000, color='b', label='b1')
ax2.plot(tw2_r.s, tw2_r.y * 1000, color='r', label='b1')
ax2.set_ylabel('y [mm]')
ax2.set_xlabel('s [mm]')

ax1.set_xlim(tw1['s', 'ip5'] - 300, tw1['s', 'ip5'] + 300)

ax1.axvline(x=tw1['s', 'ip5'], color='k', linestyle='--', alpha=0.5)
ax2.axvline(x=tw1['s', 'ip5'], color='k', linestyle='--', alpha=0.5)
plt.text(tw1['s', 'ip5'], -0.6, 'ip5', rotation=90, alpha=0.5,
         horizontalalignment='left', verticalalignment='top')

plt.show()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import time
import numpy as np

import xtrack as xt
import xpart as xp

#################################
# Load a line and build tracker #
#################################
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, p0c=7e12)
line.build_tracker()

###########################
# Generate some particles #
###########################

particles = line.build_particles(
    x_norm=np.linspace(-2, 2, 1000), y_norm=0.1, delta=3e-4,
    nemitt_x=2.5e-6, nemitt_y=2.5e-6)

p_no_optimized = particles.copy()
p_optimized = particles.copy()

##############################
# Track without optimization #
##############################

num_turns = 10
line.track(p_no_optimized, num_turns=num_turns, time=True)
t_not_optimized = line.time_last_track

####################
# Optimize tracker #
####################

line.optimize_for_tracking()

# This performs the following actions (physics model is unchanged):
# - Disables xdeps expressions
# - Removes inactive multipoles
# - Merges consecutive multipoles
# - Removes drifts with zero length
# - Merges consecutive drifts

###########################
# Track with optimization #
###########################

line.track(p_optimized, num_turns=num_turns, time=True)
t_optimized = line.time_last_track

#################
# Compare times #
#################

num_particles = len(p_no_optimized.x)
print(f'Time not optimized {t_not_optimized*1e6/num_particles/num_turns:.1f} us/part/turn')
print(f'Time optimized {t_optimized*1e6/num_particles/num_turns:.1f} us/part/turn')

###################################
# Check that result are identical #
###################################

assert np.all(p_no_optimized.state == 1)
assert np.all(p_optimized.state == 1)

assert np.allclose(p_no_optimized.x, p_optimized.x, rtol=0, atol=1e-14)
assert np.allclose(p_no_optimized.y, p_optimized.y, rtol=0, atol=1e-14)
assert np.allclose(p_no_optimized.px, p_optimized.px, rtol=0, atol=1e-14)
assert np.allclose(p_no_optimized.py, p_optimized.py, rtol=0, atol=1e-14)
assert np.allclose(p_no_optimized.zeta, p_optimized.zeta, rtol=0, atol=1e-14)
assert np.allclose(p_no_optimized.delta, p_optimized.delta, rtol=0, atol=1e-14)


import time

import xtrack as xt

from cpymad.madx import Madx

# Load the line
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
line.particle_ref = xt.Particles(p0c=7e12, mass=xt.PROTON_MASS_EV)
collider = xt.Environment(lines={'lhcb1': line})
collider.build_trackers()


tw_ref = collider.lhcb1.twiss()

ele_start_match = 's.ds.l7.b1'
ele_end_match = 'e.ds.r7.b1'
tw_init = tw_ref.get_twiss_init(ele_start_match)

ttt = collider.twiss(
    #verbose=True,
    start=[ele_start_match],
    end=[ele_end_match],
    init=tw_init,
    _keep_initial_particles=True,
    _keep_tracking_data=True,
    )

line._kill_twiss = False

n_repeat = 1000
t1 = time.perf_counter()
for repeat in range(n_repeat):
    tw = collider.twiss(
        #verbose=True,
        start=[ele_start_match],
        end=[ele_end_match],
        init=tw_init,
        _ebe_monitor=[ttt.lhcb1.tracking_data],
        _initial_particles=[ttt.lhcb1._initial_particles]
        )
t2 = time.perf_counter()

print(f'Average time: {(t2-t1)/n_repeat*1e3} ms')

line._kill_twiss = True

n_repeat = 1000
t1 = time.perf_counter()
for repeat in range(n_repeat):
    tw = collider.twiss(
        #verbose=True,
        start=[ele_start_match],
        end=[ele_end_match],
        init=tw_init,
        _ebe_monitor=[ttt.lhcb1.tracking_data],
        _initial_particles=[ttt.lhcb1._initial_particles]
        )
t2 = time.perf_counter()

print(f'Average time (with kill): {(t2-t1)/n_repeat*1e3} ms')


import xtrack as xt
import xobjects as xo

bb = xt.Bend(k0=0.001, h=0.001, length=2)
bb.integrator = 'yoshida4'
bb.num_multipole_kicks = 20

p0 = xt.Particles(x=0.0, y=0.0, delta=[0, 1e-3])

bb.model = 'bend-kick-bend'
assert bb._xobject.model == 2
p_bkb = p0.copy()
bb.track(p_bkb)

bb.model = 'rot-kick-rot'
assert bb._xobject.model == 3
p_rkr = p0.copy()
bb.track(p_rkr)

bb.model = 'mat-kick-mat'
assert bb._xobject.model == 4
p_mkm = p0.copy()
bb.track(p_mkm)

bb.model = 'drift-kick-drift-exact'
assert bb._xobject.model == 5
p_dkd1 = p0.copy()
bb.track(p_dkd1)

bb.model = 'drift-kick-drift-expanded'
assert bb._xobject.model == 6
p_dkd2 = p0.copy()
bb.track(p_dkd2)

xo.assert_allclose(p_bkb.x, p_rkr.x, rtol=0, atol=1e-12)
xo.assert_allclose(p_bkb.x, p_mkm.x, rtol=0, atol=1e-12)
xo.assert_allclose(p_bkb.x, p_dkd1.x, rtol=0, atol=1e-12)
xo.assert_allclose(p_bkb.x, p_dkd2.x, rtol=0, atol=1e-12)

xo.assert_allclose(p_bkb.px, p_rkr.px, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.px, p_mkm.px, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.px, p_dkd1.px, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.px, p_dkd2.px, rtol=0, atol=1e-14)

xo.assert_allclose(p_bkb.y, p_rkr.y, rtol=0, atol=1e-12)
xo.assert_allclose(p_bkb.y, p_mkm.y, rtol=0, atol=1e-12)
xo.assert_allclose(p_bkb.y, p_dkd1.y, rtol=0, atol=1e-12)
xo.assert_allclose(p_bkb.y, p_dkd2.y, rtol=0, atol=1e-12)

xo.assert_allclose(p_bkb.py, p_rkr.py, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.py, p_mkm.py, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.py, p_dkd1.py, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.py, p_dkd2.py, rtol=0, atol=1e-14)

xo.assert_allclose(p_bkb.zeta, p_rkr.zeta, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.zeta, p_mkm.zeta, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.zeta, p_dkd1.zeta, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.zeta, p_dkd2.zeta, rtol=0, atol=1e-14)

xo.assert_allclose(p_bkb.delta, p_rkr.delta, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.delta, p_mkm.delta, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.delta, p_dkd1.delta, rtol=0, atol=1e-14)
xo.assert_allclose(p_bkb.delta, p_dkd2.delta, rtol=0, atol=1e-14)















import numpy as np
import matplotlib.pyplot as plt
from scipy.io import loadmat

xt    = loadmat('particles_xt.mat')
rft_v = loadmat('particles_rft_volume.mat')
rft_s = loadmat('particles_rft_sbend.mat')

X = xt['x']
V = rft_v['v']
S = rft_s['s']

plt.figure(1)
plt.scatter(X[:,0]*1e3, X[:,1]*1e3, label='Xtrack',      s=100, facecolors='none', edgecolors='b')
plt.scatter(V[:,0]*1e3, V[:,1]*1e3, label='RFT::Volume', s=10, facecolors='m', edgecolors='none')
plt.scatter(S[:,0]*1e3, S[:,1]*1e3, label='RFT::SBend',  s=10, facecolors='c', edgecolors='none')
plt.xlabel("$x$ [mm]")
plt.ylabel("$p_x$ [mrad]")
plt.legend()
plt.show()

plt.figure(2)
plt.scatter(X[:,2]*1e3, X[:,3]*1e3, label='Xtrack',      s=100, facecolors='none', edgecolors='b')
plt.scatter(V[:,2]*1e3, V[:,3]*1e3, label='RFT::Volume', s=10, facecolors='m', edgecolors='none')
plt.scatter(S[:,2]*1e3, S[:,3]*1e3, label='RFT::SBend',  s=10, facecolors='c', edgecolors='none')
plt.xlabel("$y$ [mm]")
plt.ylabel("$p_y$ [mrad]")
plt.legend()
plt.show()

plt.figure(3)
plt.scatter(X[:,4]*1e3, X[:,5]*1e3, label='Xtrack',      s=100, facecolors='none', edgecolors='b')
plt.scatter(V[:,4]*1e3, V[:,5]*1e3, label='RFT::Volume', s=10, facecolors='m', edgecolors='none')
plt.scatter(S[:,4]*1e3, S[:,5]*1e3, label='RFT::SBend',  s=10, facecolors='c', edgecolors='none')
plt.xlabel("$t$ [mm/c]")
plt.ylabel("$p_t$ [mrad]")
plt.legend()
plt.show()



import pathlib

import numpy as np

import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()

@for_all_test_contexts
def test_match_coupling_knob(test_context):
    line = xt.Line.from_json(test_data_folder /
        'hllhc14_no_errors_with_coupling_knobs/line_b1.json')
    line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)

    line.build_tracker(_context=test_context)

    line.cycle('ip1', inplace=True)
    line.twiss_default['method'] = '4d'

    # Flat machine and skew quadrupoles off
    for nn in line.vars.get_table().rows['on_.*|acb.*|cm.*'].name:
        line.vars[nn] = 0

    # see Eq. 47 in https://cds.cern.ch/record/522049/files/lhc-project-report-501.pdf
    class ActionCmin(xt.Action):
        def __init__(self, line):
            self.line = line
        def run(self):
            tw = self.line.twiss()
            tt = self.line.get_table(attr=True)
            k1sl = tt['k1sl']
            c_min = 1 / (2*np.pi) * np.sum(k1sl * np.sqrt(tw.betx * tw.bety)
                                    * np.exp(1j * 2 * np.pi * (tw.mux - tw.muy)))
            return {'c_min_re': c_min.real, 'c_min_im': c_min.imag}

    act_cmin = ActionCmin(line)

    line.vars['kqs.a34b1'] = 0
    line.vars['kqs.a78b1'] = 0
    line.vars['kqs.l4b1'] = line.vars['kqs.a34b1']
    line.vars['kqs.l8b1'] = line.vars['kqs.a78b1']
    line.vars['kqs.r3b1'] = line.vars['kqs.a34b1']
    line.vars['kqs.r7b1'] = line.vars['kqs.a78b1']
    vary=xt.VaryList(['kqs.a23b1', 'kqs.a34b1', 'kqs.a67b1', 'kqs.a78b1'], step=5e-5)

    c_min_match = 1e-4

    opt_re = line.match_knob(knob_name='c_minus_re.b1',
        knob_value_start=0, knob_value_end=c_min_match,
        run=False,
        vary=vary,
        targets=[
            act_cmin.target('c_min_re', value=c_min_match, tol=1e-8),
            act_cmin.target('c_min_im', value=0, tol=1e-8),
        ])
    opt_re.solve()
    opt_re.generate_knob()

    opt_im = line.match_knob(knob_name='c_minus_im.b1',
        knob_value_start=0, knob_value_end=c_min_match,
        run=False,
        vary=vary,
        targets=[
            act_cmin.target('c_min_re', value=0, tol=1e-8),
            act_cmin.target('c_min_im', value=c_min_match, tol=1e-8),
        ])
    opt_im.solve()
    opt_im.generate_knob()

    line.vars['c_minus_re.b1'] = 5e-3
    line.vars['c_minus_im.b1'] = 0
    tt_re = line.get_table(attr=True)

    line.vars['c_minus_re.b1'] = 0
    line.vars['c_minus_im.b1'] = 5e-3
    tt_im = line.get_table(attr=True)

    # Check orthogonality
    line.vars['c_minus_re.b1'] = 1e-3
    line.vars['c_minus_im.b1'] = 1e-3
    xo.assert_allclose(line.twiss().c_minus/np.sqrt(2), 1e-3, rtol=0, atol=1.5e-5)

    # Compare against "legacy" knobs
    line_legacy = xt.Line.from_json(test_data_folder /
        'hllhc14_no_errors_with_coupling_knobs/line_b1.json')
    line_legacy.cycle('ip1', inplace=True)

    # Flat machine and skew quadrupoles off
    for nn in line_legacy.vars.get_table().rows['on_.*|acb.*|cm.*'].name:
        line_legacy.vars[nn] = 0

    tw_leg = line_legacy.twiss()

    line_legacy.vars['cmrskew'] = 5e-3
    line_legacy.vars['cmiskew'] = 0
    tt_re_leg = line_legacy.get_table(attr=True)

    line_legacy.vars['cmrskew'] = 0
    line_legacy.vars['cmiskew'] = 5e-3
    tt_im_leg = line_legacy.get_table(attr=True)

    # Withing 12% of the maximum value
    xo.assert_allclose(tt_re.k1sl, tt_re_leg.k1sl, rtol=0,
                    atol=0.12 * np.max(tt_re_leg.k1sl))
    xo.assert_allclose(tt_im.k1sl, tt_im_leg.k1sl, rtol=0,
                    atol=0.12 * np.max(tt_im_leg.k1sl))




from cpymad.madx import Madx
import xtrack as xt
import xdeps as xd

import numpy as np

orbit_settings = {
    'acbh19.r3b1': 15e-6,
    'acbv20.r3b1': 10e-6,
    'acbv19.r3b2': 15e-6,
    'acbh20.r3b2': 10e-6,
}

# Generate Xsuite maps

collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider.vars.update(orbit_settings)
collider['lhcb1'].twiss_default['method'] = '4d'
collider['lhcb2'].twiss_default['method'] = '4d'

map_b1 = xt.SecondOrderTaylorMap.from_line(
    line=collider.lhcb1, start='ip3', end='ip4')

map_b4 = xt.SecondOrderTaylorMap.from_line(
    line=collider.lhcb2, start='ip4', end='ip3')
map_b2_reflected = map_b4.scale_coordinates(scale_x=-1, scale_px=-1)

# Generate MAD-X maps

mad = Madx()
mad.input(f"""
call,file="../../test_data/hllhc15_thick/lhc.seq";
call,file="../../test_data/hllhc15_thick/hllhc_sequence.madx";
beam, sequence=lhcb1, particle=proton, pc=7000;
beam, sequence=lhcb2, particle=proton, pc=7000, bv=-1;
call,file="../../test_data/hllhc15_thick/opt_round_150_1500.madx";
""")
mad.globals.update(orbit_settings)

mad.use(sequence="lhcb1")
seq = mad.sequence.lhcb1
mad.twiss()

mad.input('''
select, flag=sectormap, pattern='ip';
twiss, sectormap, sectorpure, sectortable=secttab_b1;
''')
sectmad_b1  = xd.Table(mad.table.secttab_b1)

mad.input('''
    seqedit,sequence=lhcb2;flatten;reflect;flatten;endedit;
    use, sequence=lhcb2;
''')

mad.input('''
select, flag=sectormap, pattern='ip';
twiss, sectormap, sectorpure, sectortable=secttab_b2;
''')
sectmad_b2  = xd.Table(mad.table.secttab_b2)

# Compare

for line_name in ['lhcb1', 'lhcb2']:

    if line_name == 'lhcb1':
        xs_map = map_b1
        sectmad = sectmad_b1
        start = 'ip3'
        end = 'ip4'
        tw = collider.lhcb1.twiss()
    elif line_name == 'lhcb2':
        xs_map = map_b2_reflected
        sectmad = sectmad_b2
        start = 'ip4'
        end = 'ip3'
        tw = collider.lhcb2.twiss()
    else:
        raise ValueError(f'Unknown line_name: {line_name}')

    TT = xs_map.T
    RR = xs_map.R
    k = xs_map.k


    nemitt_x = 2.5e-6
    nemitt_y = 2.5e-6
    scale_in = [
        np.sqrt(tw['betx', start] * nemitt_x / tw['gamma0']),
        np.sqrt(tw['gamx', start] * nemitt_x / tw['gamma0']),
        np.sqrt(tw['bety', start] * nemitt_y / tw['gamma0']),
        np.sqrt(tw['gamy', start] * nemitt_y / tw['gamma0']),
        0.05,
        1e-3]

    scale_out = [
        np.sqrt(tw['betx', end] * nemitt_x / tw['gamma0']),
        np.sqrt(tw['gamx', end] * nemitt_x / tw['gamma0']),
        np.sqrt(tw['bety', end] * nemitt_y / tw['gamma0']),
        np.sqrt(tw['gamy', end] * nemitt_y / tw['gamma0']),
        0.05,
        1e-3]

    # Check k
    for ii in range(6):
        scaled_k = k[ii] / scale_out[ii]
        scaled_k_mad = sectmad[f'k{ii+1}', end] / scale_out[ii]
        # The following means that a the orbit kick is the same within 5e-5 sigmas
        assert np.isclose(scaled_k, scaled_k_mad, atol=5e-5, rtol=0)

    # Check R
    for ii in range(6):
        for jj in range(6):
            scaled_rr = RR[ii, jj] / scale_out[ii] * scale_in[jj]
            scaled_rr_mad = sectmad[f'r{ii+1}{jj+1}', end] * (
                                scale_in[jj] / scale_out[ii])
            # The following means that a change of one sigma in jj results
            # in an error of less than 5e-4 sigmas on ii
            assert np.isclose(scaled_rr, scaled_rr_mad, atol=5e-4, rtol=0)

    # Check T
    for ii in range(6):
        for jj in range(6):
            for kk in range(6):
                scaled_tt = (TT[ii, jj, kk]
                            / scale_out[ii] * scale_in[jj] * scale_in[kk])
                scaled_tt_mad = (sectmad[f't{ii+1}{jj+1}{kk+1}', end]
                                / scale_out[ii] * scale_in[jj] * scale_in[kk])
                # The following means that a change of one sigma in jj, kk results
                # in an error of less than 5e-4 sigmas on ii
                assert np.isclose(scaled_tt, scaled_tt_mad, atol=5e-4, rtol=0)






import numpy as np
from scipy.constants import c as clight
from scipy.constants import hbar
from scipy.constants import epsilon_0

from cpymad.madx import Madx
import xtrack as xt
import xobjects as xo

mad = Madx()
mad.call('../../test_data/sps_thick/sps.seq')

# mad.input('beam, particle=proton, pc=26;')
# mad.input('beam, particle=electron, pc=20;')

# # realistic
# mad.input('beam, particle=electron, pc=20;')
# v_mv = 25
# num_turns = 8000

# higher energy
mad.input('beam, particle=electron, pc=50;')
v_mv = 250
num_turns = 600

mad.call('../../test_data/sps_thick/lhc_q20.str')

mad.use(sequence='sps')

mad.input('twiss, table=tw4d;')
twm4d = mad.table.tw4d

n_cav = 6

mad.sequence.sps.elements['actcse.31632'].volt = v_mv * 10 / n_cav   # To stay in the linear region
mad.sequence.sps.elements['actcse.31632'].freq = 3
mad.sequence.sps.elements['actcse.31632'].lag = 0.5


mad.input('twiss, table=tw6d;')
twm6d = mad.table.tw6d

mad.sequence.sps.beam.radiate = True

line = xt.Line.from_madx_sequence(mad.sequence.sps, allow_thick=True,
                                  deferred_expressions=True)
line.particle_ref = xt.Particles(mass0=xt.ELECTRON_MASS_EV,
                                    q0=-1, gamma0=mad.sequence.sps.beam.gamma)
line.cycle('bpv.11706', inplace=True)

line.insert_element(element=line['actcse.31632'].copy(), index='bpv.11706',
                    name='cav1')
line.insert_element(element=line['actcse.31632'].copy(), index='bpv.21508',
                    name='cav2')
line.insert_element(element=line['actcse.31632'].copy(), index='bpv.41508',
                    name='cav4')
line.insert_element(element=line['actcse.31632'].copy(), index='bpv.51508',
                    name='cav5')
line.insert_element(element=line['actcse.31632'].copy(), index='bpv.61508',
                    name='cav6')
mad.emit()

tt = line.get_table()

# Remove edge effects
# for nn in tt.rows[tt.element_type=='DipoleEdge'].name:
#     line[nn].k = 0

tw_thick = line.twiss()

Strategy = xt.slicing.Strategy
Teapot = xt.slicing.Teapot

line.discard_tracker()
slicing_strategies = [
    Strategy(slicing=Teapot(1)),  # Default
    Strategy(slicing=Teapot(2), element_type=xt.Bend),
    Strategy(slicing=Teapot(8), element_type=xt.Quadrupole),
]

line.slice_thick_elements(slicing_strategies)
line.build_tracker()

line.discard_tracker()
line.build_tracker(_context=xo.ContextCpu())
line.configure_radiation(model=None)

match_chrom = True

# line.vars['klsda'] = 0.0
# line.vars['klsdb'] = 0.0
# line.vars['klsfa'] = 0.0
# line.vars['klsfb'] = 0.0
# line.vars['klsfc'] = 0.0
# match_chrom = False

line.vars['mdv.52907.ksl0'] = 0.0
line.vars['mdv.53107.ksl0'] = 0.0
line.vars['mdv.53307.ksl0'] = 0.0
line.vars['mdv.53507.ksl0'] = 0.0

line.element_refs['mdv.52907'].ksl[0] = line.vars['mdv.52907.ksl0']
line.element_refs['mdv.53107'].ksl[0] = line.vars['mdv.53107.ksl0']
line.element_refs['mdv.53307'].ksl[0] = line.vars['mdv.53307.ksl0']
line.element_refs['mdv.53507'].ksl[0] = line.vars['mdv.53507.ksl0']

# Kill sextupoles in the bump
line.element_refs['lsf.53205'].k2l = 0
line.element_refs['lsd.53505'].k2l = 0
line.element_refs['lsf.53605'].k2l = 0
line.element_refs['lsd.60105'].k2l = 0

tw0 = line.twiss()
opt_bump = line.match(
    solve=False,
    method='4d',
    start='mdv.52707', end='mdv.60107',
    init_at=xt.START, init=tw0,
    vary=[
        xt.VaryList(['mdv.52907.ksl0', 'mdv.53107.ksl0',
                     'mdv.53307.ksl0', 'mdv.53507.ksl0'], step=1e-5),
    ],
    targets=[
        xt.Target(lambda tw: np.max(tw.y), value=20e-3),
        xt.TargetSet(y=0, py=0, at=xt.END)
    ],
)
opt_bump.solve()

opt = line.match(
    solve=False,
    vary=[
        xt.VaryList(['kqf', 'kqd'], step=1e-5, tag='tune'),
        xt.VaryList(['klsda', 'klsdb', 'klsfa', 'klsfb', 'klsfc'], step=1e-4, tag='chrom'),
    ],
    targets=[
        xt.TargetSet(qx=20.13, qy=20.18, tol=1e-7, tag='tune'),
        xt.TargetSet(dqx=2.0, dqy=2.0, tol=1, tag='chrom'),
        ],
)

opt.enable_all_targets()
opt.enable_all_vary()
opt.disable_targets(tag='chrom')
opt.disable_vary(tag='chrom')
opt.solve()

if match_chrom:

    opt.disable_all_targets()
    opt.disable_all_vary()
    opt.enable_targets(tag='chrom')
    opt.enable_vary(tag='chrom')
    opt.solve()

    opt.enable_all_targets()
    opt.enable_all_vary()
    opt.solve()


tw = line.twiss(strengths=True)
tw4d = line.twiss(method='4d')

line.configure_radiation(model='mean')

# Tapering!!!
line.compensate_radiation_energy_loss()

tw_rad = line.twiss(eneloss_and_damping=True, method='6d',
                    use_full_inverse=False)
tw_rad2 = line.twiss(eneloss_and_damping=True, method='6d',
                     radiation_method='full')


assert tw_rad.eq_gemitt_x is not None
assert tw_rad.eq_gemitt_y is not None
assert tw_rad.eq_gemitt_zeta is not None

ex = tw_rad.eq_nemitt_x / (tw_rad.gamma0 * tw_rad.beta0)
ey = tw_rad.eq_nemitt_y / (tw_rad.gamma0 * tw_rad.beta0)
ez = tw_rad.eq_nemitt_zeta / (tw_rad.gamma0 * tw_rad.beta0)

line.configure_radiation(model='quantum')

p = line.build_particles(num_particles=1000)
line.discard_tracker()
line.build_tracker(_context=xo.ContextCpu(omp_num_threads='auto'), use_prebuilt_kernels=False)
line.track(p, num_turns=num_turns, time=True, turn_by_turn_monitor=True)
print(f'Tracking time: {line.time_last_track}')

twe = tw.rows[:-1]
hl = twe.angle_rad
dl = twe.length
hh = hl * 0
hh[dl>0] = hl[dl>0] / dl[dl>0]
gamma0 = tw.gamma0
cur_H_x = twe.gamx * twe.dx**2 + 2 * twe.alfx * twe.dx * twe.dpx + twe.betx * twe.dpx**2
I5_x  = np.sum(cur_H_x * hh**3 * dl)
I2_x = np.sum(hh**2 * dl)
I4_x = np.sum(twe.dx * hh**3 * dl) # to be generalized for combined function magnets

cur_H_y = twe.gamy * twe.dy**2 + 2 * twe.alfy * twe.dy * twe.dpy + twe.bety * twe.dpy**2
I5_y  = np.sum(cur_H_y * hh**3 * dl)
I2_y = np.sum(hh**2 * dl)
I4_y = np.sum(twe.dy * hh**3 * dl) # to be generalized for combined function magnets

lam_comp = 2.436e-12 # [m]
ex_hof = 55 * np.sqrt(3) / 96 * lam_comp / 2 / np.pi * gamma0**2 * I5_x / (I2_x - I4_x)
ey_hof = 55 * np.sqrt(3) / 96 * lam_comp / 2 / np.pi * gamma0**2 * I5_y / (I2_y - I4_y)

mon = line.record_last_track

sigma_tab = tw_rad.get_beam_covariance(gemitt_x=tw_rad.eq_gemitt_x,
                                       gemitt_y=tw_rad.eq_gemitt_y,
                                       gemitt_zeta=tw_rad.eq_gemitt_zeta)
sigma_betatron_tab = tw_rad.get_beam_covariance(gemitt_x=tw_rad.eq_gemitt_x,
                                                gemitt_y=tw_rad.eq_gemitt_y,
                                                gemitt_zeta=0)

import matplotlib.pyplot as plt
plt.close('all')
fig = plt.figure(1, figsize=(6.4, 4.8 * 1.8))

spo = fig. add_subplot(4, 1, 1)
spo.plot(tw.s, tw.y)
spo.plot(tw_rad.s, tw_rad.y, '--')

plt.xlabel('s [m]')
plt.ylabel('y [m]')

spx = fig. add_subplot(4, 1, 2)
spx.plot(np.std(mon.x, axis=0))
spx.axhline(sigma_tab.sigma_x[0], color='red')
spx.axhline(tw_rad2.eq_beam_covariance_matrix.sigma_x[0], color='green')
plt.ylabel(r'$\sigma_x$ [m]')
# spx.axhline(np.sqrt(ex_hof * tw.betx[0] + (np.std(p.delta) * tw.dx[0])**2), color='green')

spy = fig. add_subplot(4, 1, 3, sharex=spx)
spy.plot(np.std(mon.y, axis=0))
spy.axhline(sigma_tab.sigma_y[0], color='red')
spy.axhline(tw_rad2.eq_beam_covariance_matrix.sigma_y[0], color='green')
plt.ylim(0, 3e-4)
plt.ylabel(r'$\sigma_y$ [m]')
# spy.axhline(np.sqrt(ey_hof * tw.bety[0] + (np.std(p.delta) * tw.dy[0])**2), color='green')

spz = fig. add_subplot(4, 1, 4, sharex=spx)
spz.plot(np.std(mon.zeta, axis=0))
spz.axhline(sigma_tab.sigma_zeta[0], color='red')
spz.axhline(tw_rad2.eq_beam_covariance_matrix.sigma_zeta[0], color='green')

plt.ylabel(r'$\sigma_z$ [m]')
plt.xlabel('Turns')
plt.suptitle(f"Qx = {tw.qx:.2f} - Qy = {tw.qy:.2f} - Q'x = {tw.dqx:.2f} - Q'y = {tw.dqy:.2f}")
plt.subplots_adjust(left=.16, hspace=.37)

# plt.savefig(f'sps_eq_emitt_dqx_{tw.dqx:.2f}_dqy_{tw.dqy:.2f}_freq.png', dpi=300)

# Plot horizontal and vertical dispersions

fig = plt.figure(2, figsize=(6.4, 4.8 * 1.2))
spdx = fig.add_subplot(2, 1, 1)
spdx.plot(tw.s, tw.dx)
plt.ylabel(r'$D_x$ [m]')

spdy = fig.add_subplot(2, 1, 2, sharex=spdx)
spdy.plot(tw.s, tw.dy)

plt.xlabel('s [m]')
plt.ylabel(r'$D_y$ [m]')

mad.input('''
mdv.52907, kick := mdv.52907.ksl0;
mdv.53107, kick := mdv.53107.ksl0;
mdv.53307, kick := mdv.53307.ksl0;
mdv.53507, kick := mdv.53507.ksl0;
''')


import xtrack as xt
import xobjects as xo
import numpy as np

env = xt.Environment()

line = env.new_line(
    components=[
        env.new('q0', 'Quadrupole', length=2.0, at=20.0),
        env.new('ql', 'Quadrupole', length=2.0, at=10.0),
        env.new('qr', 'Quadrupole', length=2.0, at=30),
        env.new('end', 'Marker', at=50.),
    ])


line.insert([
    env.new('q4', 'q0', anchor='center', at=0, from_='q0@end'), # will replace half of q0
    env.new('q5', 'q0', at=0, from_='ql'), # will replace the full ql
    env.new('m5.0', 'Marker', at='q5@start'),
    env.new('m5.1', 'Marker', at='q5@start'),
    env.new('m5.2', 'Marker', at='q5@end'),
    env.new('m5.3', 'Marker'),
])

line.insert([
    env.new('q6', 'q0', at=0, from_='qr'),
    env.new('mr.0', 'Marker', at='qr@start'),
    env.new('mr.1', 'Marker', at='qr@start'),
    env.new('mr.2', 'Marker', at='qr@end'),
    env.new('mr.3', 'Marker'),
])

tt = line.get_table()
tt.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt.name == np.array(
    ['drift_1', 'm5.0', 'm5.1', 'q5', 'm5.2', 'm5.3', 'drift_2',
       'q0_entry', 'q0..0', 'q4', 'drift_3..1', 'mr.0', 'mr.1', 'q6',
       'mr.2', 'mr.3', 'drift_4', 'end', '_end_point']))
xo.assert_allclose(tt.s_center, np.array(
    np.array([ 4.5,  9. ,  9. , 10. , 11. , 11. , 15. , 19. , 19.5, 21. , 25.5,
       29. , 29. , 30. , 31. , 31. , 40.5, 50. , 50.])),
    rtol=0., atol=1e-14)


import xtrack as xt
import xobjects as xo
import numpy as np

line = xt.Line(elements={
    's1': xt.Sextupole(k2=0.1, length=9.0)
})
line.particle_ref = xt.Particles(p0c=7000e9)

line_1slice = line.copy(shallow=True)
line_1slice.slice_thick_elements(
    slicing_strategies=[
        xt.Strategy(slicing=xt.Teapot(1), element_type=xt.Sextupole),
    ])

line_3slices = line.copy(shallow=True)
line_3slices.slice_thick_elements(
    slicing_strategies=[
        xt.Strategy(slicing=xt.Teapot(3), element_type=xt.Sextupole),
    ])

tw_1slice = line_1slice.twiss(betx=1, bety=1, x=1e-2)
tw_3slices = line_3slices.twiss(betx=1, bety=1, x=1e-2)

tw_1kick = line.twiss(betx=1, bety=1, x=1e-2)

line.configure_sextupole_model(num_multipole_kicks=3)

assert line['s1'].num_multipole_kicks == 3
tw_3kicks = line.twiss(betx=1, bety=1, x=1e-2)

assert np.abs(tw_1slice.x[-1] - tw_3slices.x[-1]) > 1e-6

xo.assert_allclose(tw_1slice.s[-1], tw_1kick.s[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_1slice.x[-1], tw_1kick.x[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_1slice.px[-1], tw_1kick.px[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_1slice.y[-1], tw_1kick.y[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_1slice.py[-1], tw_1kick.py[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_1slice.zeta[-1], tw_1kick.zeta[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_1slice.ptau[-1], tw_1kick.ptau[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_1slice.betx[-1], tw_1kick.betx[-1], atol=1e-10, rtol=0)
xo.assert_allclose(tw_1slice.bety[-1], tw_1kick.bety[-1], atol=1e-10, rtol=0)

xo.assert_allclose(tw_3slices.s[-1], tw_3kicks.s[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_3slices.x[-1], tw_3kicks.x[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_3slices.px[-1], tw_3kicks.px[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_3slices.y[-1], tw_3kicks.y[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_3slices.py[-1], tw_3kicks.py[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_3slices.zeta[-1], tw_3kicks.zeta[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_3slices.ptau[-1], tw_3kicks.ptau[-1], atol=1e-15, rtol=0)
xo.assert_allclose(tw_3slices.betx[-1], tw_3kicks.betx[-1], atol=1e-10, rtol=0)
xo.assert_allclose(tw_3slices.bety[-1], tw_3kicks.bety[-1], atol=1e-10, rtol=0)
